import logging
import os
import re

import netmiko
import yaml

from dcim.models import Switch
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import transaction

from inventory.models import L2Address, InventoryObject

log = logging.getLogger('importers.switches')


class MatchLinesError(Exception):
    def __init__(self, lines, expected_lines):
        self.lines = lines
        self.expected_lines = expected_lines

        super(MatchLinesError, self).__init__("{!r} != {!r}".format(lines, expected_lines))


def match_lines(lines, expected_lines):
    if not _match_lines(lines, expected_lines):
        raise MatchLinesError(lines, expected_lines)


def _match_lines(lines, expected_lines):
    for line, expected_line in zip(lines, expected_lines):
        if hasattr(expected_line, 'match'):
            if not expected_line.match(line):
                return False
        else:
            if expected_line != line:
                return False

    return True


def get_table_body(lines, expected_header=None, expected_footer=None):
    if expected_header is None:
        expected_header = []

    if expected_footer is None:
        expected_footer = []

    header = lines[:len(expected_header)]

    match_lines(header, expected_header)

    if len(expected_footer):
        footer = lines[-len(expected_footer):]
        body = lines[len(header):-len(footer)]

        match_lines(footer, expected_footer)
    else:
        # lines[:-0] and lines[-0:] would work improperly
        body = lines[len(header):]

    return body


class SwitchHandler(object):
    def __init__(self, switch, credentials):
        self.address = switch.management_address
        self.username = credentials['username']
        self.password = credentials['password']

    def get_mac_to_port_mapping(self):
        mapping = list(self.mac_to_port_iter())

        for mac_address, port in mapping:
            assert L2Address.is_valid_mac_address(mac_address)

        return mapping


class HPProcurveHandler(SwitchHandler):
    def mac_to_port_iter(self):
        net_connect = netmiko.ConnectHandler(device_type='hp_procurve', ip=self.address,
                                             username=self.username, password=self.password,
                                             verbose=False)

        output = net_connect.send_command('show mac-address')

        lines = output.splitlines()

        body = get_table_body(lines, [
            '',
            ' Status and Counters - Port Address Table',
            '',
            '  MAC Address   Located on Port',
            '  ------------- ---------------',
        ], [
            ' '
        ])

        for line in body:
            raw_mac_address, raw_port = line.strip().split()

            mac_address = raw_mac_address.replace('-', '')
            port = int(raw_port)

            yield mac_address, port


class CiscoIOSHandler(SwitchHandler):
    def mac_to_port_iter(self):
        net_connect = netmiko.ConnectHandler(device_type='cisco_ios', ip=self.address,
                                             username=self.username, password=self.password,
                                             verbose=False)

        output = net_connect.send_command('show mac address-table dynamic')

        lines = output.splitlines()

        body = get_table_body(lines, [
            '          Mac Address Table',
            '-------------------------------------------',
            '',
            'Vlan    Mac Address       Type        Ports',
            '----    -----------       --------    -----'
        ], [
            re.compile(r'Total Mac Addresses for this criterion: \d+$')
        ])

        for line in body:
            vlan, raw_mac_address, type_, port_name = line.strip().split()

            if type_ == 'DYNAMIC':
                mac_address = raw_mac_address.replace('.', '')

                yield mac_address, port_name


HANDLERS = {
    'hp-procurve': HPProcurveHandler,
    'cisco-ios': CiscoIOSHandler,
}


def mac_to_port(switch, all_credentials):
    credentials = all_credentials.get(switch.label)

    if not credentials:
        log.warning("No credentials supplied for %s", switch)
        return []

    try:
        switch_type = credentials['type']
    except KeyError:
        log.error("No switch type specified for %s", switch)
        return []

    handler_class = HANDLERS.get(switch_type)

    if not handler_class:
        log.warning("Unknown switch type %s", switch_type)
        return []

    try:
        handler = handler_class(switch, credentials)
        return handler.get_mac_to_port_mapping()
    except:
        log.error("Failed import from switch %s as %s", switch, switch_type, exc_info=True)
        return []


def load_credentials():
    path = settings.NETWORK_SWITCHES_CREDENTIALS_PATH
    conf_dirs = settings.SITE_CONF_DIRS

    for conf_dir in conf_dirs:
        full_path = os.path.join(conf_dir, path)

        if os.path.exists(full_path):
            with open(full_path) as f:
                return yaml.safe_load(f)

    raise ImproperlyConfigured("Credentials file %r not found in %r" % (path, conf_dirs))


def import_address_tables():
    log.info('Start importing mac addresses from switches')

    credentials = load_credentials()

    switches = Switch.objects.all()

    for switch in switches:
        log.info('Importing from switch %s at %s', switch, switch.management_address)

        for mac_address, port_name in mac_to_port(switch, credentials):
            if not L2Address.is_valid_mac_address(mac_address):
                log.error('Invalid mac address %r', mac_address)
                continue

            l2address, created = L2Address.objects.get_or_create(address=mac_address)

            switch_port, created = switch.ports.get_or_create(switch_port_name=port_name)

            switch_port.l2addresses.add(l2address)

    log.info('Done importing mac addresses from switches')


def bind_dcim_servers_to_inventory():
    for l2address in L2Address.objects.filter(inventory_object__isnull=False).\
            select_related('inventory_object'):
        ports_with_address_and_server = l2address.switch_ports.filter(dcim_server_id__isnull=False)

        if ports_with_address_and_server:
            # Server mac address should be bound to only one endpoint port (endpoint port is one
            # connected to server)
            assert len(ports_with_address_and_server) == 1

            port = ports_with_address_and_server[0]

            inventory_object_at_port = InventoryObject.objects.get(dcim_id=port.dcim_server_id)
            inventory_object_at_address = l2address.inventory_object

            if inventory_object_at_port != inventory_object_at_address:
                inventory_object_at_address.dcim_id = inventory_object_at_port.dcim_id
                inventory_object_at_address.dcim_label = inventory_object_at_port.dcim_label

                with transaction.atomic():
                    # First orphan object at port, to avoid unique constraint conflict
                    inventory_object_at_port.dcim_orphan()
                    # Save bound object
                    inventory_object_at_address.save()

    InventoryObject.prune_orphaned_objects()


PORT_LABEL_RE = re.compile("""
    ([a-z]+)  # interface type
    \s*
    (
        (?: \d+/)*
        (?: \d+)
    )  # port number
""", re.VERBOSE | re.IGNORECASE)


def get_port_name_variants(label):
    yield label.strip()

    prefix = 'Port'
    if label.startswith(prefix):
        yield label[len(prefix):].strip()

    m = PORT_LABEL_RE.match(label)
    if m:
        interface_type, port_number = m.groups()
        yield interface_type[:2] + port_number


def bind_dcim_to_switch_ports():
    for switch in Switch.objects.all():
        by_label = {}
        by_switch_port_name = {}

        for port in switch.ports.all():
            if port.label:
                by_label[port.label] = port

            if port.switch_port_name:
                by_switch_port_name[port.switch_port_name] = port

        for label, port in by_label.items():
            for port_name in get_port_name_variants(label):
                if not port_name:
                    continue

                switch_port = by_switch_port_name.get(port_name)

                if switch_port and switch_port != port:
                    port.switch_port_name = switch_port.switch_port_name
                    for l2address in switch_port.l2addresses.all():
                        port.l2addresses.add(l2address)
                    port.save()
                    switch_port.delete()

                    log.info("Bound label %s to name %s at %s", label, port_name, switch)
                    break
