from __future__ import absolute_import

import logging
from functools import wraps

from celery import shared_task, chain, Task
from django.conf import settings

import dcim.importer
import importers.ldap
import importers.switches
import importers.zabbix
import importers.puppet


log = logging.getLogger('importers.tasks')


def racks_task(func):
    @shared_task(bind=True, ignore_result=True)
    @wraps(func)
    def task(self, *args, **kwargs):
        log.info("Starting task %s...", self.name)
        try:
            func(*args, **kwargs)
        except Exception, e:
            log.exception("Task %s failed: %s.", self.name, e)
        else:
            log.info("Task %s done.", self.name)

    task.time_limit = settings.TASK_TIME_LIMITS[task.name]

    return task


@racks_task
def import_zabbix_hosts():
    importers.zabbix.import_hosts()


@racks_task
def import_puppet_attributes():
    importers.puppet.import_puppet_attributes()


@racks_task
def import_dcim_servers():
    dcim.importer.import_servers()


@racks_task
def import_dcim_switches():
    dcim.importer.import_switches()


@racks_task
def import_switches_address_tables():
    importers.switches.import_address_tables()


@racks_task
def bind_dcim_to_switch_ports():
    importers.switches.bind_dcim_to_switch_ports()


@racks_task
def bind_dcim_servers_to_inventory():
    importers.switches.bind_dcim_servers_to_inventory()


@racks_task
def import_users_from_ldap():
    importers.ldap.import_users()


@shared_task
def _chain_task_finish(name):
    log.info("Chain task %s finished." % name)


class ChainTask(Task):
    abstract = True

    def __init__(self):
        subtasks = [
            task.si() for task in self.subtasks
        ]

        subtasks.append(_chain_task_finish.si(self.name))

        self.chain = chain(*subtasks)

    def apply(self, *args, **kwargs):
        return self.chain.apply(*args, **kwargs)

    def apply_async(self, *args, **kwargs):
        return self.chain.apply_async(*args, **kwargs)


@shared_task
class import_all(ChainTask):
    subtasks = (
        import_zabbix_hosts,
        import_puppet_attributes,
        import_dcim_servers,
        import_dcim_switches,
        import_switches_address_tables,
        bind_dcim_to_switch_ports,
        bind_dcim_servers_to_inventory,
        import_users_from_ldap,
    )
