import json

import mockldap
from django.conf import settings
from django.test import TestCase
from mock import patch, call

import dcim.importer
import importers.ldap
import importers.puppet
import importers.switches
import importers.zabbix
from dcim.models import Switch
from inventory.models import (
    FacterLocation,
    InventoryObject,
    InventoryObjectType,
    L2Address,
    Lab,
    RacksUser,
    RacksGroup,
)


class ZabbixTestCase(TestCase):
    @patch('pyzabbix.ZabbixAPI')
    def test(self, mock_class):
        instance = mock_class.return_value
        instance.hostgroup.get.return_value = [
            {
                'groupid': 1
            }
        ]
        instance.host.get.return_value = [
            {
                'host': 'hostname1',
                'hostid': 2
            }
        ]

        importers.zabbix.import_hosts()

        self.assertItemsEqual(InventoryObject.objects.values('name'), [{'name': 'hostname1'}])


class MockResponse(object):
    pass


class PuppetTestCase(TestCase):
    maxDiff = None

    def setUp(self):
        self.inventory_object = InventoryObject.objects.create(id=1, name='object1')
        # This object should be ignored by importer
        InventoryObject.objects.create(id=2, name=None)

    def assert_attributes(self, inventory_object, attributes):
        self.assertItemsEqual(inventory_object.attributes.all().values('name', 'value'), [
            {
                'name': key,
                'value': value,
            }
            for key, value in attributes.items()
        ])

    @patch('requests.get')
    def run_import(self, response, mock_function):
        mock_response = mock_function.return_value = MockResponse()
        mock_response.text = json.dumps(response)

        importers.puppet.import_puppet_attributes()

        self.assertItemsEqual(mock_function.mock_calls, [
            call(u'testpuppet/puppet/node/object1?environment=production',
                 cert=('/etc/ssl/certs/puppet.crt', '/etc/ssl/private/puppet.key'),
                 verify=False,
                 headers={'User-agent': 'Racks Puppet Importer/0.2'})
        ])

    def test_hw_server(self):
        self.inventory_object.update_attributes({
            'to_change': 'original_value',
        })

        self.run_import({
            'data': {
                'parameters': {
                    'virtual': 'physical',
                    'attribute1': 'value1',
                    'mtu_docker0': 'ignored_attribute_value',
                    'to_change': 'changed_value',
                }
            }
        })

        object1 = InventoryObject.objects.get(id=1)

        self.assertEqual(object1.type.name, 'HW Server')
        self.assert_attributes(object1, {
            'attribute1': 'value1',
            'virtual': 'physical',
            'to_change': 'changed_value',
        })

    def test_kvm_server(self):
        self.inventory_object.update_attributes({
            'to_change': 'original_value',
        })

        self.run_import({
            'data': {
                'parameters': {
                    'virtual': 'kvm',
                    'attribute1': 'value1',
                    'mtu_docker0': 'ignored_attribute_value',
                    'to_change': 'changed_value',
                }
            }
        })

        object1 = InventoryObject.objects.get(id=1)

        self.assertEqual(object1.type.name, 'KVM Virtual Machine')
        self.assert_attributes(object1, {
            'attribute1': 'value1',
            'virtual': 'kvm',
            'to_change': 'changed_value',
        })


class DCIMTestCase(TestCase):
    @patch('dcim.importer.Session.request')
    def test_import_servers(self, mock_function):
        mock_function.return_value = [
            {
                'DeviceID': 1,
                'Label': 'label1'
            },
            {
                'DeviceID': 3,
                'Label': 'label3'
            },
        ]

        # This label will be changed)
        InventoryObject.objects.create(id=1, dcim_id=1, dcim_label='oldLabel1')
        # This will be orphaned and pruned, as it's not in DCIM
        InventoryObject.objects.create(id=2, dcim_id=2, dcim_label='oldLabel2')

        # Not orphaned
        # Has name
        InventoryObject.objects.create(id=4, dcim_id=4, dcim_label='label4', name='name4')
        # Has attribute
        obj5 = InventoryObject.objects.create(id=5, dcim_id=5, dcim_label='label5')
        obj5.attributes.create(name='name1', value='value1')
        # Has type
        obj6 = InventoryObject(id=6, dcim_id=6, dcim_label='label6')
        obj6.type = InventoryObjectType.objects.create(name='type1')
        obj6.save()
        # Has location
        obj7 = InventoryObject(id=7, dcim_id=7, dcim_label='label7')
        obj7.facter_location = FacterLocation.objects.create()
        obj7.save()
        # Has label
        obj8 = InventoryObject.objects.create(id=8, dcim_id=8, dcim_label='label8')
        obj8.label.create(name='objectLabel1')
        # Has l2network
        obj9 = InventoryObject.objects.create(id=9, dcim_id=9, dcim_label='label9')
        obj9.l2network.create(name='l2network1', vlanid=1)
        # Has ipv4network
        obj10 = InventoryObject.objects.create(id=10, dcim_id=10, dcim_label='label10')
        obj10.ipv4network.create()
        # Has responsible
        obj11 = InventoryObject.objects.create(id=11, dcim_id=11, dcim_label='label11')
        obj11.responsible.create(username='user1')
        # Has l2address
        obj12 = InventoryObject.objects.create(id=12, dcim_id=12, dcim_label='label12')
        obj12.l2addresses.create(address='112233445566')

        dcim.importer.import_servers()

        self.assertItemsEqual(
            InventoryObject.objects.values_list('id', 'dcim_id', 'dcim_label', 'orphan_dcim_id'),
            [
                (1, 1, 'label1', None),
                (13, 3, 'label3', None),
                # Not pruned
                (4, None, 'label4', 4),
                (5, None, 'label5', 5),
                (6, None, 'label6', 6),
                (7, None, 'label7', 7),
                (8, None, 'label8', 8),
                (9, None, 'label9', 9),
                (10, None, 'label10', 10),
                (11, None, 'label11', 11),
                (12, None, 'label12', 12),
            ])

    @patch('dcim.importer.Session.request')
    def test_import_switches(self, mock_function):
        def side_effect(object_type, object_id=None, filters={}):
            if object_type == 'device':
                return [
                    {
                        'DeviceID': 1,
                        'PrimaryIP': 'ip1',
                        'Label': 'label1',
                    },
                    {
                        'DeviceID': 3,
                        'PrimaryIP': 'ip3',
                        'Label': 'label3',
                    },
                ]

            if object_type == 'deviceport':
                if object_id == 1:
                    return {
                        '1': {
                            'Label': 'portLabel1',
                            'ConnectedDeviceID': None,
                        },
                        '2': {
                            'Label': 'portLabel2',
                            'ConnectedDeviceID': None,
                        },
                        '3': {
                            'Label': 'portLabel3',
                            'ConnectedDeviceID': '4',
                        },
                        '4': {
                            'Label': 'portLabel4',
                            'ConnectedDeviceID': '9999999',
                        },
                    }

                return {}

        mock_function.side_effect = side_effect

        # Inventory object for port3
        InventoryObject.objects.create(dcim_id=4)

        # This switch will have label changed
        switch1 = Switch.objects.create(dcim_id=1, label='oldLabel1', management_address='oldIp1')

        switch1.ports.create(number=2, label='portOldLabel2')
        switch1.ports.create(number=3, label='portLabel3', dcim_server_id=1)
        switch1.ports.create(number=4, label='portLabel4', dcim_server_id=1)
        # This port will get deleted
        switch1.ports.create(number=5, label='portLabel5')

        # This switch will get deleted
        Switch.objects.create(dcim_id=2, label='label2'),

        dcim.importer.import_switches()

        self.assertItemsEqual(
            Switch.objects.values('label', 'management_address'),
            [
                {'label': 'label1', 'management_address': 'ip1'},
                {'label': 'label3', 'management_address': 'ip3'},
            ]
        )

        self.assertItemsEqual(
            Switch.objects.get(label='label1').ports.values('number', 'label', 'dcim_server_id'),
            [
                # New port
                {'number': '1', 'label': 'portLabel1', 'dcim_server_id': None},
                # Label changed
                {'number': '2', 'label': 'portLabel2', 'dcim_server_id': None},
                # dcim_server changed
                {'number': '3', 'label': 'portLabel3', 'dcim_server_id': 4},
                # dcim_server changed to none because no corresponding server exist
                {'number': '4', 'label': 'portLabel4', 'dcim_server_id': None},
            ]
        )


class SwitchesTestCase(TestCase):
    @patch('importers.switches.load_credentials')
    @patch('importers.switches.mac_to_port')
    def test_import_address_tables(self, mock_mac_to_port, mock_load_credentials):
        mock_mac_to_port.return_value = [
            # invalid address, won't get added
            ('xxxxx', 'Gi0/1'),
            ('aaaaaa000001', 'Gi0/2'),
        ]

        Switch.objects.create(dcim_id=1, label='ip1', management_address='ip1')

        importers.switches.import_address_tables()

        self.assertItemsEqual(
            L2Address.objects.values('address'),
            [
                {'address': 'aaaaaa000001'},
            ]
        )

        addr1 = L2Address.objects.get(address='aaaaaa000001')

        self.assertItemsEqual(addr1.switch_ports.values_list('switch_port_name'), [('Gi0/2',)])

    def test_bind_dcim_to_switch_ports(self):
        switch1 = Switch.objects.create(dcim_id=1, label='ip1', management_address='ip1')

        # This should stay intact
        switch1.ports.create(id=1, label='GigabitEthernet0/1', switch_port_name='Gi0/1')
        port2 = switch1.ports.create(id=2, label='GigabitEthernet0/2')
        port2.l2addresses.create(address='aaaaaa000001')
        # This should get merged to port2
        port3 = switch1.ports.create(id=3, switch_port_name='Gi0/2')
        port3.l2addresses.create(address='aaaaaa000002')

        importers.switches.bind_dcim_to_switch_ports()

        self.assertItemsEqual(switch1.ports.values_list('id', 'label', 'switch_port_name'), [
            (1, 'GigabitEthernet0/1', 'Gi0/1'),
            (2, 'GigabitEthernet0/2', 'Gi0/2'),
        ])

        self.assertItemsEqual(port2.l2addresses.values_list('address'), [
            ('aaaaaa000001',),
            ('aaaaaa000002',),
        ])

    def test_bind_dcim_servers_to_inventory(self):
        obj1 = InventoryObject.objects.create(name='obj1')
        # This will get merged into obj1 and pruned
        obj2 = InventoryObject.objects.create(dcim_id=2)

        l2address1 = obj1.l2addresses.create(address='aaaaaa000001')

        switch3 = Switch.objects.create(dcim_id=3)
        port1 = switch3.ports.create(dcim_server_id=obj2.dcim_id)

        port1.l2addresses.add(l2address1)

        importers.switches.bind_dcim_servers_to_inventory()

        self.assertItemsEqual(
            InventoryObject.objects.values_list('name', 'dcim_id'),
            [
                # obj2 bound to obj1
                ('obj1', 2),
            ]
        )


class LDAPImportTestCase(TestCase):
    DIRECTORY = {
        settings.LDAP_USER_SEARCH.base_dn: {
            'ou': ['people']
        },
        settings.AUTH_LDAP_USER_DN_TEMPLATE % {'user': 'user1'}: {
            'uid': ['user1'],
            'objectClass': ['person'],
            'givenName': ['Given1'],
            'mail': ['user1@mail'],
            'sn': ['sn1'],
        },
        settings.AUTH_LDAP_GROUP_SEARCH.base_dn: {
            'cn': ['group1'],
            'objectClass': ['groupOfNames'],
            'member': [settings.AUTH_LDAP_USER_DN_TEMPLATE % {'user': 'user1'}],
        }
    }

    print DIRECTORY

    @classmethod
    def setUpClass(cls):
        cls.mockldap = mockldap.MockLdap(cls.DIRECTORY)

    @classmethod
    def tearDownClass(cls):
        del cls.mockldap

    def setUp(self):
        self.mockldap.start()
        self.ldapobj = self.mockldap[settings.AUTH_LDAP_SERVER_URI]

    def tearDown(self):
        # Stop patching ldap.initialize and reset state.
        self.mockldap.stop()
        del self.ldapobj

    def test_import_users(self):
        # This user should be deleted
        RacksUser.objects.create(username='user2')
        # This group should be deleted
        RacksGroup.objects.create(name='group2')

        # This users should be marked as deleted
        responsible_object_user = RacksUser.objects.create(username='responsible_object_user')
        inventory_object = InventoryObject.objects.create()
        inventory_object.responsible.add(responsible_object_user)
        inventory_object.assigned_user = RacksUser.objects.create(username='assigned_object_user')

        responsible_lab_user = RacksUser.objects.create(username='responsible_lab_user')
        lab = Lab.objects.create()
        lab.responsible.add(responsible_lab_user)
        lab.assigned_user = RacksUser.objects.create(username='assigned_lab_user')

        # This groups should be marked deleted
        inventory_object.assigned_group = RacksGroup.objects.create(name='assigned_object_group')
        lab.assigned_group = RacksGroup.objects.create(name='assigned_lab_group')

        inventory_object.save()
        lab.save()

        importers.ldap.import_users()

        self.assertEqual(RacksUser.objects.count(), 5)

        self.assertFalse(RacksUser.objects.get(username='user1').deleted)
        self.assertTrue(RacksUser.objects.get(username='responsible_object_user').deleted)
        self.assertTrue(RacksUser.objects.get(username='responsible_lab_user').deleted)
        self.assertTrue(RacksUser.objects.get(username='assigned_object_user').deleted)
        self.assertTrue(RacksUser.objects.get(username='assigned_lab_user').deleted)

        self.assertEqual(RacksGroup.objects.count(), 3)
        self.assertFalse(RacksGroup.objects.get(name='group1').deleted)
        self.assertTrue(RacksGroup.objects.get(name='assigned_object_group').deleted)
        self.assertTrue(RacksGroup.objects.get(name='assigned_lab_group').deleted)

        user1 = RacksUser.objects.get(username='user1')

        self.assertEqual(user1.first_name, 'Given1')
        self.assertEqual(user1.last_name, 'sn1')
        self.assertEqual(user1.email, 'user1@mail')
        self.assertEqual([g.name for g in user1.groups.all()], ['group1'])
