from django import forms
from django.forms.util import flatatt
from django.utils.datastructures import MultiValueDict, MergeDict
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from inventory.models import (
    InventoryObject,
    InventoryObjectLabel
)
from inventory.widgets import FilteredSelectMultiple


class TagsInput(forms.Widget):
    def render(self, name, value, attrs=None):
        if value is None:
            value = ''
        final_attrs = self.build_attrs(attrs, name=name)

        if final_attrs.get('readonly'):
            output = [
                '<div class="form-control" readonly="readonly">'
            ]
            for text in value:
                output.append(format_html('<span class="label label-default">{0}</span>',
                                          text))
            output.append('</div>')
        else:
            final_attrs['data-role'] = 'tagsinput'
            output = [
                format_html('<select multiple="multiple"{0}>', flatatt(final_attrs))
            ]
            for text in value:
                output.append(format_html('<option value="{0}">{1}</option>', text, text))

        output.append('</select>')
        return mark_safe('\n'.join(output))

    def value_from_datadict(self, data, files, name):
        if isinstance(data, (MultiValueDict, MergeDict)):
            return data.getlist(name)
        return data.get(name, None)


class TagsField(forms.Field):
    widget = TagsInput

    def prepare_value(self, value):
        if value is None:
            return None

        if isinstance(value, self.BoundData):
            return value.data

        objects = InventoryObjectLabel.objects.filter(pk__in=value)
        return [obj.name for obj in objects]

    class BoundData:
        def __init__(self, data):
            self.data = data

    def bound_data(self, data, initial):
        return self.BoundData(data)

    def to_python(self, value):
        objects = []

        for name in value:
            obj, created = InventoryObjectLabel.objects.get_or_create(name=name)
            objects.append(obj)

        return objects


class InventoryObjectForm(forms.ModelForm):
    label = TagsField(required=None)

    class Meta:
        model = InventoryObject
        widgets = {
            'responsible': FilteredSelectMultiple,
            'l2network': FilteredSelectMultiple,
            'ipv4network': FilteredSelectMultiple,
            'bugs': forms.TextInput(attrs={'data-role': 'bugsinput'})
        }
