import netaddr
import re
import urlparse

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.http import urlencode

from inventory.fields import NullableCharField
from security import permissions


class CreateTimeBaseObject(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    active = models.BooleanField(default=True, db_index=True, editable=False)
    added_by = models.CharField(max_length=64, null=True, blank=True, editable=False)

    class Meta:
        abstract = True


class Location(CreateTimeBaseObject):
    name = models.CharField(max_length=254, unique=True)

    def __unicode__(self):
        return self.name

    class Meta:
        ordering = ('name',)


class FacterLocation(CreateTimeBaseObject):
    fact = models.CharField(max_length=32, unique=True)
    location = models.ForeignKey('Location', on_delete=models.SET_NULL, related_name='facts',
                                 null=True, blank=True)

    def __unicode__(self):
        return "{} ({})".format(self.fact, self.location)


class L2Network(CreateTimeBaseObject):
    name = models.CharField(max_length=254)
    vlanid = models.IntegerField(validators=[
        MinValueValidator(1),
        MaxValueValidator(4094),
    ])

    class Meta:
        verbose_name = 'VLAN'

    def __unicode__(self):
        if self.name:
            return '{} (VLAN ID: {})'.format(self.name, self.vlanid)

        return "VLAN ID: {}".format(self.vlanid)


class L2Address(models.Model):
    MAC_ADDRESS_RE = re.compile(r'^[0-9a-f]{12}$')

    address = models.CharField(max_length=12, unique=True)
    inventory_object = models.ForeignKey('inventory.InventoryObject', null=True,
                                         related_name='l2addresses')

    @classmethod
    def is_valid_mac_address(cls, mac_address):
        return cls.MAC_ADDRESS_RE.match(mac_address)

    def __unicode__(self):
        return ":".join(self.address[i:i + 2] for i in range(0, len(self.address), 2))


class IPv4Network(CreateTimeBaseObject):
    address = models.CharField(max_length=15)
    netmask = models.CharField(max_length=15)
    gateway = models.CharField(max_length=15)
    l2network = models.ForeignKey(
        L2Network, on_delete=models.SET_NULL, related_name='ipv4_networks', null=True, blank=True)

    class Meta:
        unique_together = ('address', 'netmask')
        ordering = ('address',)
        verbose_name = 'IPv4 Network'

    def __str__(self):
        return netaddr.IPNetwork('%s/%s' % (self.address, self.netmask))

    def __unicode__(self):
        return netaddr.IPNetwork('%s/%s' % (self.address, self.netmask))


class InventoryObjectType(CreateTimeBaseObject):
    name = models.CharField(max_length=254, unique=True)

    def __str__(self):
        return self.name

    def __unicode__(self):
        return self.name


class InventoryObjectLabel(CreateTimeBaseObject):
    TYPE_CHOICES = (
        ('GY', 'Grey'),
        ('BL', 'Blue'),
        ('GN', 'Green'),
        ('LB', 'Lightblue'),
        ('OR', 'Orange'),
        ('RD', 'Red'),
    )

    name = models.CharField(max_length=64, unique=True)
    type = models.CharField(
        max_length=2,
        choices=TYPE_CHOICES,
        default='GY'
    )

    def __unicode__(self):
        return self.name

    def __str__(self):
        return self.name

    def get_bootstrap_color(self):
        colors = {
            'GY': 'default',
            'BL': 'primary',
            'GN': 'success',
            'LB': 'info',
            'OR': 'warning',
            'RD': 'danger',
        }

        return colors.get(self.type, 'default')

    def get_absolute_url(self):
        return reverse('inventory.objects.list') + '?' + urlencode({
            'label': self.name
        })


class InventoryObject(CreateTimeBaseObject):
    name = NullableCharField(max_length=254, unique=True, db_index=True, null=True, blank=True)
    type = models.ForeignKey(
        InventoryObjectType, related_name='inventory_objects', null=True, blank=True)
    facter_location = models.ForeignKey('FacterLocation', related_name='inventory_objects',
                                        null=True, blank=True)
    label = models.ManyToManyField(InventoryObjectLabel, related_name='inventory_objects',
                                   blank=True)
    l2network = models.ManyToManyField(L2Network, related_name='inventory_objects', null=True,
                                       blank=True)
    ipv4network = models.ManyToManyField(IPv4Network, related_name='inventory_objects', null=True,
                                         blank=True)
    responsible = models.ManyToManyField('RacksUser', related_name='inventory_objects',
                                         db_index=True, blank=True)

    ipmi_host = NullableCharField(max_length=255, unique=True, null=True, blank=True)
    ipmi_user = NullableCharField(max_length=255, null=True, blank=True)
    ipmi_password = NullableCharField(max_length=255, null=True, blank=True)

    dcim_id = models.IntegerField(null=True, db_index=True, unique=True, blank=True)
    orphan_dcim_id = models.IntegerField(null=True, db_index=True, blank=True)
    dcim_label = NullableCharField(max_length=255, null=True, blank=True)
    dcim_jira_issue = models.CharField(max_length=255, null=True, blank=True)

    bugs = models.CommaSeparatedIntegerField(blank=True, null=True, max_length=254)

    assigned_user = models.ForeignKey('RacksUser', blank=True, null=True,
                                      related_name='assigned_objects')
    assigned_group = models.ForeignKey('RacksGroup', blank=True, null=True,
                                       related_name='assigned_objects')

    class Meta:
        ordering = ('-created',)

    def __unicode__(self):
        return "{} ({})".format(self.name, self.dcim_label)

    def update_attributes(self, attributes):
        for name, value in attributes.items():
            attribute_object, created = self.attributes.get_or_create(name=name,
                                                                      defaults={'value': value})

            if not created and attribute_object.value != value:
                attribute_object.value = value
                attribute_object.save()

    @property
    def cpu(self):
        try:
            return '%sx%s' % (
                self.attributes.get(name='processorcount').value,
                self.attributes.get(name='processor0').value)
        except ObjectDoesNotExist:
            return None

    @property
    def display_name(self):
        if self.name:
            return self.name
        elif self.dcim_label:
            return 'DCIM Entry: %s' % (self.dcim_label, )
        elif self.dcim_id:
            return 'DCIM Entry ID: %s' % (self.dcim_id, )
        elif self.orphan_dcim_id:
            return 'Orphaned DCIM Entry ID: %s' % (self.orphan_dcim_id, )
        else:
            return 'Empty object ID: %s' % (self.id, )

    def dump(self):
        return {
            'name': self.name,
            'hardware': self.hardware,
        }

    @property
    def hardware(self):
        try:
            return self.attributes.get(name='hardware').value
        except ObjectDoesNotExist:
            return None

    @property
    def ipmi_link(self):
        if self.ipmi_host:
            if urlparse.urlparse(self.ipmi_host).scheme != '':
                return self.ipmi_host
            return 'http://%s' % self.ipmi_host

        return None

    @property
    def ipv4address(self):
        try:
            return self.attributes.get(name='ipaddress').value
        except ObjectDoesNotExist:
            return None

    @property
    def ipv4networks(self):
        return [
            '%s/%s' % (x.address, x.netmask) for x in self.ipv4network.all()]

    @property
    def jira_issue(self):
        try:
            return settings.JIRA_BROWSE_URL + self.attributes.get(
                name='jira_issue').value
        except:
            return None

    @property
    def dcim_jira_issue_link(self):
        if self.dcim_jira_issue:
            return settings.JIRA_BROWSE_URL + self.dcim_jira_issue

    @property
    def macaddress(self):
        try:
            return [
                x.value for x in self.attributes.filter(
                    name__contains='macaddress')]
        except ObjectDoesNotExist:
            return []

    @property
    def memory(self):
        try:
            return self.attributes.get(name='memorysize').value
        except ObjectDoesNotExist:
            return None

    @property
    def ssh_link(self):
        try:
            return 'ssh://%s' % self.attributes.get(name='fqdn').value
        except ObjectDoesNotExist:
            return None

    def get_absolute_url(self):
        return reverse('inventory.objects.detail', args=(self.pk,))

    @property
    def label_string(self):
        return ','.join([x.name for x in self.label.all()])

    @property
    def macaddress_string(self):
        return ','.join(self.macaddress)

    @property
    def responsible_string(self):
        return ','.join([x.username for x in self.responsible.all()])

    @property
    def zabbix_link(self):
        try:
            zabbix_url = 'https://monitoring.infra.mirantis.net'
            return '%s/hostinventories.php?hostid=%s' % (
                zabbix_url,
                self.attributes.get(name='zabbix_host_id').value)
        except ObjectDoesNotExist:
            return None

    @property
    def dcim_link(self):
        if self.dcim_id:
            return settings.DCIM_URL + 'devices.php?' + urlencode({'DeviceID': self.dcim_id})

    @property
    def location_name(self):
        if self.facter_location and self.facter_location.location:
            return self.facter_location.location.name

        return None

    @property
    def location_fact(self):
        if self.facter_location:
            return self.facter_location.fact

        return None

    def dcim_orphan(self):
        self.orphan_dcim_id = self.dcim_id
        self.dcim_id = None
        self.save()

    @classmethod
    def prune_orphaned_objects(cls):
        """Remove object which were merged into other and have no related entries"""

        orphaned_objects = cls.objects.filter(name=None,
                                              type=None,
                                              facter_location=None,
                                              label=None,
                                              l2network=None,
                                              ipv4network=None,
                                              responsible=None,
                                              attributes=None,
                                              l2addresses=None,
                                              dcim_id=None)

        orphaned_objects.delete()

    @property
    def bugs_list(self):
        return self.bugs.split(',') if self.bugs else []

    PUBLIC_FIELDS = (
        'id',
        'created',
        'modified',
        'name',
        'type',
        'facter_location',
        'label',
        'l2network',
        'ipv4network',
        'responsible',
        # 'ipmi_host',
        # 'ipmi_user',
        # 'ipmi_password',
        'dcim_id',
        'orphan_dcim_id',
        'dcim_label',
        'bugs',
        'assigned_user',
        'assigned_group',
        'cpu',
        'display_name',
        'hardware',
        # 'ipmi_link',
        'ipv4address',
        'ipv4networks',
        'jira_issue',
        'macaddress',
        'memory',
        'ssh_link',
        'zabbix_link',
        'dcim_link',
        'location_name',
        'location_fact',
        'bugs_list',
        'attributes',
        'l2addresses',
    )

    @classmethod
    def get_permissions(cls, user, obj):
        if obj is not None and user is not None and obj.can_view_private(user):
            return permissions.VIEW_ALL

        return permissions.Permissions(view=cls.PUBLIC_FIELDS)

    def can_view_private(self, user):
        if self.assigned_user == user:
            return True

        if self.assigned_group in user.groups.all():
            return True

        for lab in self.labs.all():
            if lab.assigned_user == user:
                return True

            if lab.assigned_group in user.groups.all():
                return True

        return False

    @classmethod
    def get_queryset_for_permissions(cls, qs, user):
        permissions_by_object = {}

        for obj in qs.prefetch_related('assigned_user', 'assigned_group', 'labs',
                                       'labs__assigned_user', 'labs__assigned_group'):
            permissions_by_object[obj] = obj.get_permissions(user)

        return permissions_by_object


class InventoryObjectCredential(CreateTimeBaseObject):
    name = models.CharField(max_length=255)
    context = models.CharField(max_length=255)
    value = models.CharField(max_length=255)
    inventory_object = models.ForeignKey(
        InventoryObject, related_name='inventory_objects')

    class Meta:
        unique_together = ('name', 'context', 'inventory_object')

    def __unicode__(self):
        return '%s for %s@%s' % (
            self.name, self.context, self.inventory_object.name)


class InventoryObjectAttribute(CreateTimeBaseObject):
    name = models.CharField(max_length=254)
    value = models.TextField()
    object = models.ForeignKey(InventoryObject, related_name='attributes')

    class Meta:
        unique_together = ('name', 'object')
        ordering = ('name',)

    def __unicode__(self):
        return '%s@%s' % (self.name, self.object.name)

    def __str__(self):
        return '%s@%s' % (self.name, self.object.name)

    def awesome_value(self):
        handlers = [
            'fqdn',
            'puppet_manifests_version',
            'zabbix_host_id',
        ]
        if self.name in handlers:
            if self.name == 'zabbix_host_id':
                return '<a href="%s" target="_blank">%s</a>' % (
                    self.object.zabbix_link,
                    self.value
                )
            if self.name == 'puppet_manifests_version':
                sha1commit = re.search(r'([a-f0-9]{40})', self.value)
                if sha1commit and len(sha1commit.groups()) > 0:
                    return '<a href="%s:%s" target="_blank">%s</a>' % (
                        'https://review.fuel-infra.org/#/q/commit',
                        sha1commit.groups()[0],
                        self.value,
                    )
                return self.value
            if self.name == 'fqdn':
                return '<a href="ssh://%s">%s</a>' % (
                    self.value,
                    self.value,
                )

        return self.value


class Lab(CreateTimeBaseObject):
    name = models.CharField(max_length=255, blank=False)
    vlans = models.ManyToManyField('L2Network', null=True, blank=True)
    responsible = models.ManyToManyField('RacksUser', blank=True, related_name='labs')
    inventory_objects = models.ManyToManyField('InventoryObject', related_name='labs', blank=True)
    bugs = models.CommaSeparatedIntegerField(blank=True, null=True, max_length=254)
    assigned_user = models.ForeignKey('RacksUser', blank=True, null=True,
                                      related_name='assigned_labs')
    assigned_group = models.ForeignKey('RacksGroup', blank=True, null=True,
                                       related_name='assigned_labs')

    @property
    def bugs_list(self):
        return self.bugs.split(',') if self.bugs else []


class RacksUser(models.Model):
    username = models.CharField(max_length=255)
    first_name = models.CharField(max_length=255)
    last_name = models.CharField(max_length=255)
    email = models.EmailField(blank=True)
    deleted = models.BooleanField(default=False)

    def __unicode__(self):
        if self.deleted:
            return '%s (deleted)' % self.username

        return self.username

    def prune(self):
        """Remove user if it is not associated with any object"""

        if (
            not self.inventory_objects.exists() and
            not self.labs.exists() and
            not self.assigned_objects.exists() and
            not self.assigned_labs.exists()
        ):
            self.delete()
        else:
            self.deleted = True
            self.save()


class RacksGroup(models.Model):
    name = models.CharField(max_length=255)
    deleted = models.BooleanField(default=False)

    users = models.ManyToManyField('RacksUser', related_name='groups', blank=True)

    def __unicode__(self):
        if self.deleted:
            return '%s (deleted)' % self.name

        return self.name

    def prune(self):
        """Remove group if it is not associated with any object"""
        if (
            not self.assigned_objects.exists() and
            not self.assigned_labs.exists()
        ):
            self.delete()
        else:
            self.deleted = True
            self.save()
