from rest_framework import serializers

from inventory.models import (
    FacterLocation,
    InventoryObject,
    InventoryObjectAttribute,
    InventoryObjectLabel,
    InventoryObjectType,
    IPv4Network,
    RacksUser,
)
from security.permissions import get_permissions


HIDDEN_VALUE = '__access_denied__'


class AccessDenied(Exception):
    pass


class ProtectedObjectWrapper(object):
    """
    This wrapper raises AccessDenied exception on attempt to get forbidden attribute and ignore
    setting of protected attribute
    """

    def __init__(self, obj, request):
        permissions = get_permissions(request, obj._meta.model, obj)

        # Can't use simple double underscore here, because member name mangling does not work with
        # setattr
        object.__setattr__(self, '_ProtectedObjectWrapper__obj', obj)
        object.__setattr__(self, '_ProtectedObjectWrapper__permissions', permissions)

    def __getattr__(self, name):
        value = getattr(self.__obj, name)

        # This attributes are used internally by rest serializer
        if name in ('_meta', 'full_clean', 'save'):
            return value

        field_permissions = self.__permissions.get_field_permissions(name)

        if not field_permissions.view:
            raise AccessDenied

        return value

    def __setattr__(self, name, value):
        # This attributes are used internally by rest serializer
        if name in ('_related_data', '_m2m_data', '_nested_forward_relations'):
            object.__setattr__(self, name, value)
            return

        field_permissions = self.__permissions.get_field_permissions(name)

        if field_permissions.edit:
            setattr(self.__obj, name, value)

    def __delattr__(self, name):
        if name in ('_related_data', '_m2m_data', '_nested_forward_relations'):
            object.__delattr__(self, name)
            return

        delattr(self.__obj, name)


class ProtectedFieldWrapper(object):
    """
    This wrapper catches AccessDenied exception on serialization and returns HIDDEN_VALUE instead
    """

    def __init__(self, field):
        self.__field = field

    def field_to_native(self, obj, field_name):
        try:
            return self.__field.field_to_native(obj, field_name)
        except AccessDenied:
            return HIDDEN_VALUE

    def __getattr__(self, name):
        return getattr(self.__field, name)


class ProtectedModelWrapper(object):
    """This wrapper return protected object wrapper on model instantiation"""

    def __init__(self, parent, model):
        self.__model = model
        self.__parent = parent

    def __call__(self, *args, **kwargs):
        request = self.__parent.context['request']

        return ProtectedObjectWrapper(self.__model(*args, **kwargs), request)

    def __getattr__(self, name):
        return getattr(self.__model, name)


class ProtectedModelSerializer(serializers.ModelSerializer):
    """
    This serializer replaces fields and model with wrappers and passes protected object to
    serializer and deserializer methods.
    """

    def __init__(self, *args, **kwargs):
        super(ProtectedModelSerializer, self).__init__(*args, **kwargs)

        self.fields = serializers.SortedDict((field_name, ProtectedFieldWrapper(field))
                                             for field_name, field in self.fields.items())

        self.opts.model = ProtectedModelWrapper(self, self.opts.model)

    def to_native(self, obj):
        # If obj is None, or already wrapped (as happens in update view I think), don't wrap it
        # (again)
        if obj is None or isinstance(obj, ProtectedObjectWrapper):
            return super(ProtectedModelSerializer, self).to_native(obj)

        protected_object = ProtectedObjectWrapper(obj, self.context['request'])

        return super(ProtectedModelSerializer, self).to_native(protected_object)

    def restore_object(self, attrs, instance=None):
        if instance is not None:
            instance = ProtectedObjectWrapper(instance, self.context['request'])

        return super(ProtectedModelSerializer, self).restore_object(attrs, instance)


class FacterLocationSerializer(ProtectedModelSerializer):
    class Meta:
        model = FacterLocation
        fields = ('fact',)


class InventoryObjectTypeSerializer(ProtectedModelSerializer):
    class Meta:
        model = InventoryObjectType
        fields = ('name',)


class UserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = RacksUser
        fields = ('id', 'url', 'username', 'email', 'groups',)


class IPv4NetworkSerializer(ProtectedModelSerializer):
    class Meta:
        model = IPv4Network
        fields = ('id', 'address', 'netmask', 'gateway',)


class InventoryObjectLabelSerializer(ProtectedModelSerializer):
    class Meta:
        model = InventoryObjectLabel


class InventoryObjectAttributeSerializer(ProtectedModelSerializer):
    class Meta:
        model = InventoryObjectAttribute
        fields = ('id', 'object', 'name', 'value')


class RelatedListField(serializers.Field):
    def __init__(self, model_field, lookup):
        self.model_field = model_field
        self.lookup = lookup

        super(RelatedListField, self).__init__()

    def field_to_native(self, obj, field_name):
        return [getattr(x, self.lookup) for x in getattr(obj, self.model_field).all()]

    def field_from_native(self, data, files, field_name, into):
        values = data.get(field_name)
        manager = getattr(self.parent.opts.model, self.model_field).field.rel.to._default_manager

        if values is not None:
            value_objects = []
            for value in values:
                value_obj, created = manager.get_or_create(**{self.lookup: value})
                value_objects.append(value_obj)
            into[self.model_field] = value_objects


class InventoryObjectSerializer(ProtectedModelSerializer):
    labels = RelatedListField('label', 'name')
    location = serializers.SerializerMethodField('get_location')
    facter_location = FacterLocationSerializer(many=False, required=False)
    type = InventoryObjectTypeSerializer(many=False, required=False)
    attributes = serializers.SerializerMethodField('attributes_dict')
    responsible = RelatedListField('responsible', 'username')

    class Meta:
        model = InventoryObject
        fields = (
            'id',
            'name',
            'location',
            'facter_location',
            'type',
            'labels',
            'attributes',
            'responsible',)

    def get_location(self, obj):
        if obj.facter_location and obj.facter_location.location:
            return {
                'name': obj.facter_location.location.name
            }

        return None

    def attributes_dict(self, obj):
        attrs = {}
        for attr in obj.attributes.all():
            attrs[attr.name] = attr.value

        return attrs
