from __future__ import unicode_literals

import json

from django.contrib.auth.models import User
from mock import patch
from rest_framework.test import APIClient, APITestCase

from inventory.models import (
    FacterLocation,
    InventoryObject,
    InventoryObjectAttribute,
    InventoryObjectLabel,
    InventoryObjectType,
    IPv4Network,
    RacksUser,
)
from security.models import Application
from security.permissions import Permissions


class APIv2TestCase(APITestCase):
    maxDiff = None

    def setUp(self):
        user1 = User.objects.create_superuser(username='user1', password='password1',
                                              email='user1@mail')

        self.client.force_authenticate(user=user1)

        # TODO: group serialization does not work yet
        # group1 = Group.objects.create()
        # user1.groups.add(group1)

        Application.objects.create(user=user1, name='authapp1',
                                   auth_token='authtoken1', active=True)

        Application.objects.create(user=user1, name='authapp2',
                                   auth_token='authtoken2', active=True)

        location1 = FacterLocation.objects.create(
            fact='location1_shortname',
        )
        location2 = FacterLocation.objects.create(
            fact='location2_shortname',
        )

        IPv4Network.objects.create(address='10.0.0.1',
                                   netmask='255.0.0.0',
                                   gateway='10.0.0.254')

        type1 = InventoryObjectType.objects.create(name='type1')

        label1 = InventoryObjectLabel.objects.create(name='label1')

        obj1 = InventoryObject.objects.create(
            name='object1',
            facter_location=location1,
            type=type1,
        )

        obj1.attributes.create(name='attribute1',
                               value='obj1 attribute1 value')
        obj1.attributes.create(name='attribute2',
                               value='obj1 attribute2 value')

        obj2 = InventoryObject.objects.create(name='object2',
                                              facter_location=location2)
        obj2.attributes.create(name='attribute1',
                               value='obj2 attribute1 value')
        obj2.responsible.create(username='user1', email='user1@mail')

        obj2.label.add(label1)

    def test_app_authentication(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 200)

    def test_user_authentication(self):
        client = APIClient()

        User.objects.create_superuser('admin', 'foo@foo.com', 'admin')
        client.login(username='admin', password='admin')

        response = client.post('/api/v2/objects/', {
            'name': 'object3',
        })

        self.assertEqual(response.status_code, 201)

    def test_authenticated_create(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.post('/api/v2/objects/', {
            'name': 'object3',
        })

        self.assertEqual(response.status_code, 201)

    def test_invalid_auth_app(self):
        authenticated_client = APIClient(
            HTTP_X_AUTH_APPLICATION='invalidauthapp',
            HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def test_invalid_auth_token(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='invalidauthtoken')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def test_inactive_app(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp2',
                                         HTTP_X_AUTH_TOKEN='authtoken2')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 200)

        Application.objects.filter(name='authapp2').update(active=False)

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def assertResponseJSONContains(self, response, status_code, expected):
        if response.status_code != status_code:
            message = '{} != {}\n{}'.format(response.status_code, status_code,
                                            response)

            self.fail(message)

        response_json = response.rendered_content
        actual = json.loads(response_json)

        if isinstance(expected, list):
            self.assertIsInstance(actual, list)
            expected_items = expected
            actual_items = actual
        else:
            expected_items = [expected]
            actual_items = [actual]

        # To account for additional fields, which may appear in responses,
        # delete fields which aren't present in expected result before
        # comparison.
        # Can't use assertDictContainsSubset here, because we can have several
        # items here and I want to have diff of full response on test failure.
        for actual_item, expected_item in zip(actual_items, expected_items):
            if isinstance(expected_item, dict) and isinstance(actual_item,
                                                              dict):
                extra_keys = set(actual_item) - set(expected_item)

                for key in extra_keys:
                    del actual_item[key]

        self.assertEqual(actual, expected)

    def test_get_objects(self):
        response = self.client.get('/api/v2/objects/?ordering=id')

        self.assertResponseJSONContains(response, 200, [
            {
                'id': 1,
                'name': 'object1',
                'facter_location': {'fact': 'location1_shortname'},
                'type': {'name': 'type1'},
                'labels': [],
                'attributes': {
                    'attribute1': 'obj1 attribute1 value',
                    'attribute2': 'obj1 attribute2 value',
                },
                'responsible': []
            },
            {
                'id': 2,
                'name': 'object2',
                'facter_location': {'fact': 'location2_shortname'},
                'type': None,
                'labels': ['label1'],
                'attributes': {
                    'attribute1': 'obj2 attribute1 value'
                },
                'responsible': ['user1']
            },
        ])

    def test_create_object(self):
        response = self.client.post('/api/v2/objects/', {
            'name': 'object3',
        })

        self.assertResponseJSONContains(response, 201, {
            'id': 3,
            'name': 'object3',
            'facter_location': None,
            'type': None,
            'labels': [],
            'attributes': {},
            'responsible': []
        })

    def test_update_object(self):
        response = self.client.patch('/api/v2/objects/1/', {
            'name': 'modified_object1',
        })

        self.assertEqual(response.status_code, 200)
        self.assertEqual(InventoryObject.objects.get(id=1).name, 'modified_object1')

    def test_create_object_with_attributes(self):
        object_response = self.client.post('/api/v2/objects/', {
            'name': 'object3',
            'attributes': {
                'attribute1': 'obj3 attribute1 value',
            }
        })

        self.assertResponseJSONContains(object_response, 201, {
            'id': 3,
            'name': 'object3',
            'facter_location': None,
            'type': None,
            'labels': [],
            'attributes': {},
            'responsible': []
        })

        object_id = object_response.data['id']

        attribute_response = self.client.post('/api/v2/attributes/', {
            'object': object_id,
            'name': 'attribute1',
            'value': 'obj3 attribute1 value',
        })

        self.assertResponseJSONContains(attribute_response, 201, {
            'object': object_id,
            'name': 'attribute1',
            'value': 'obj3 attribute1 value',
        })

        get_response = self.client.get('/api/v2/objects/{}/'.format(object_id))

        self.assertResponseJSONContains(get_response, 200, {
            'id': 3,
            'name': 'object3',
            'facter_location': None,
            'type': None,
            'labels': [],
            'attributes': {
                'attribute1': 'obj3 attribute1 value',
            },
            'responsible': []
        })

    def test_delete_attribute(self):
        get_response = self.client.get('/api/v2/attributes/', {
            'object': 1,
            'name': 'attribute1',
        })

        self.assertResponseJSONContains(get_response, 200, [
            {'name': 'attribute1', 'value': 'obj1 attribute1 value'},
        ])

        attribute_id = get_response.data[0]['id']

        delete_response = self.client.delete(
            '/api/v2/attributes/{}/'.format(attribute_id))

        self.assertEqual(delete_response.status_code, 204)

        response = self.client.get('/api/v2/objects/1/')

        self.assertResponseJSONContains(response, 200, {
            'id': 1,
            'attributes': {
                'attribute2': 'obj1 attribute2 value',
            },
            'responsible': []
        })

    def test_get_label_by_name(self):
        response = self.client.get('/api/v2/labels/', {'name': 'label1'})

        self.assertResponseJSONContains(response, 200, [
            {'id': 1, 'name': 'label1'}
        ])

    def test_create_label_with_defaults(self):
        create_response = self.client.post('/api/v2/labels/',
                                           {'name': 'label2'})

        self.assertResponseJSONContains(create_response, 201,
                                        {'id': 2, 'name': 'label2'})

        response = self.client.get('/api/v2/labels/')

        self.assertResponseJSONContains(response, 200, [
            {'id': 1, 'name': 'label1', 'type': 'GY'},
            {'id': 2, 'name': 'label2', 'type': 'GY'},
        ])

    def test_create_label(self):
        create_response = self.client.post('/api/v2/labels/', {
            'name': 'label3',
            'type': 'BL',
        })

        self.assertResponseJSONContains(
            create_response, 201, {'id': 2, 'name': 'label3', 'type': 'BL'})

        response = self.client.get('/api/v2/labels/')

        self.assertResponseJSONContains(response, 200, [
            {'id': 1, 'name': 'label1', 'type': 'GY'},
            {'id': 2, 'name': 'label3', 'type': 'BL'},
        ])

    def test_create_conflicting_label(self):
        create_response = self.client.post('/api/v2/labels/',
                                           {'name': 'label1'})

        self.assertResponseJSONContains(create_response, 400, {
            'name': ['Inventory object label with this Name already exists.']})

    def test_create_user_with_defaults(self):
        create_response = self.client.post('/api/v2/users/',
                                           {'username': 'user2'})

        self.assertResponseJSONContains(create_response, 201, {
            'id': 2,
            'username': 'user2',
            'url': 'http://testserver/api/v2/users/2/',
            'email': '',
            'groups': []
        })

        response = self.client.get('/api/v2/users/?ordering=id')

        self.assertResponseJSONContains(response, 200, [
            {'id': 1, 'username': 'user1'},
            {
                'id': 2,
                'username': 'user2',
                'url': 'http://testserver/api/v2/users/2/',
                'email': '',
            },
        ])

    def test_create_user(self):
        create_response = self.client.post('/api/v2/users/', {
            'username': 'user3',
            'email': 'user3@mail.nonexistent',
        })

        self.assertResponseJSONContains(create_response, 201, {
            'id': 2,
            'username': 'user3',
            'url': 'http://testserver/api/v2/users/2/',
            'email': 'user3@mail.nonexistent',
        })

        response = self.client.get('/api/v2/users/?ordering=id')

        self.assertResponseJSONContains(response, 200, [
            {'id': 1, 'username': 'user1'},
            {
                'id': 2,
                'username': 'user3',
                'url': 'http://testserver/api/v2/users/2/',
                'email': 'user3@mail.nonexistent',
            },
        ])

    def test_objects_name_filter(self):
        response = self.client.get('/api/v2/objects/?name=object2')

        self.assertResponseJSONContains(response, 200, [
            {'id': 2},
        ])

    def test_objects_location_shortname_filter(self):
        response = self.client.get(
            '/api/v2/objects/?facter_location__fact=location1_shortname')

        self.assertResponseJSONContains(response, 200, [
            {'id': 1},
        ])

    def test_object_type_name_filter(self):
        response = self.client.get(
            '/api/v2/objects/?type__name=type1')

        self.assertResponseJSONContains(response, 200, [
            {'id': 1},
        ])

    def test_get_users(self):
        response = self.client.get('/api/v2/users/')

        self.assertResponseJSONContains(response, 200, [
            {
                'url': 'http://testserver/api/v2/users/1/',
                'username': 'user1',
                'email': 'user1@mail',
                'groups': [
                ],
            },
        ])

    def test_get_ipv4networks(self):
        response = self.client.get('/api/v2/ipv4networks/')

        self.assertResponseJSONContains(response, 200, [
            {
                'id': 1,
                'address': '10.0.0.1',
                'netmask': '255.0.0.0',
                'gateway': '10.0.0.254',
            }
        ])

    def test_patch_object_labels_format_json(self):
        # Due to bug in implementation, request only works with format='json'
        patch_response = self.client.patch('/api/v2/objects/2/', {
            'labels': ['newlabel1', 'newlabel2', 'newlabel3']
        }, format='json')

        self.assertEqual(patch_response.status_code, 200)

        response = self.client.get('/api/v2/objects/2/')

        self.assertResponseJSONContains(response, 200, {
            'id': 2,
            'name': 'object2',
            'labels': ['newlabel1', 'newlabel2', 'newlabel3'],
        })

    def test_patch_object_responsible_format_json(self):
        # Due to bug in implementation, request only works with format='json'
        RacksUser.objects.create(username='user2', email='user2@mail')
        RacksUser.objects.create(username='user3', email='user3@mail')

        patch_response = self.client.patch('/api/v2/objects/2/', {
            'responsible': ['user2', 'user3']
        }, format='json')

        self.assertEqual(patch_response.status_code, 200)

        get_response = self.client.get('/api/v2/objects/2/')

        self.assertResponseJSONContains(get_response, 200, {
            'id': 2,
            'name': 'object2',
            'responsible': ['user2', 'user3']
        })


class UnprivilegedUserTestCase(APITestCase):
    maxDiff = None

    object_permissions_mock = patch(
        'inventory.models.InventoryObject.get_permissions',
        return_value=Permissions(view=('id', 'name', 'label', 'attributes'), edit=('name', 'label'))
    )
    object_permissions_mock_no_edit_labels = patch(
        'inventory.models.InventoryObject.get_permissions',
        return_value=Permissions(view=('id', 'name', 'label', 'attributes'), edit=('name',))
    )
    attribute_permissions_mock = patch(
        'inventory.models.InventoryObjectAttribute.get_permissions',
        return_value=Permissions(edit=('value',)), create=True)

    def setUp(self):
        user1 = User.objects.create_user(username='user1', password='password1',
                                         email='user1@mail')

        self.client.force_authenticate(user=user1)

        obj1 = InventoryObject.objects.create(
            id=1,
            name='object1',

        )

        obj1.attributes.create(id=1, name='attribute1',
                               value='obj1 attribute1 value')
        obj1.attributes.create(name='attribute2',
                               value='obj1 attribute2 value')

        label1 = InventoryObjectLabel.objects.create(name='label1')
        obj1.label.add(label1)

    def test_hidden_values(self):
        with self.object_permissions_mock:
            response = self.client.get('/api/v2/objects/1/')

            self.assertEqual(response.status_code, 200)

            self.assertEqual(response.data, {
                "id": 1,
                "name": "object1",
                "location": "__access_denied__",
                "facter_location": "__access_denied__",
                "type": "__access_denied__",
                "labels": ["label1"],
                "attributes": {
                    "attribute1": "obj1 attribute1 value",
                    "attribute2": "obj1 attribute2 value",
                },
                "responsible": "__access_denied__",
            })

    def test_partial_update(self):
        with self.attribute_permissions_mock:
            response = self.client.patch('/api/v2/attributes/1/', {
                'name': 'newname',
                'value': 'newvalue',
            })

            self.assertEqual(response.status_code, 200)

            attr1 = InventoryObjectAttribute.objects.get(id=1)

            # name should not be changed, because it's not editable
            self.assertEqual(attr1.name, 'attribute1')
            self.assertEqual(attr1.value, 'newvalue')

    def test_related_field_partial_update(self):
        with self.object_permissions_mock:
            response = self.client.patch('/api/v2/objects/1/', {
                'name': 'newname',
                'labels': ['newlabel1', 'newlabel2'],
            }, format='json')

            self.assertEqual(response.status_code, 200)

            obj1 = InventoryObject.objects.get(id=1)

            self.assertEqual(obj1.name, 'newname')

            self.assertItemsEqual(obj1.label.values_list('name', flat=True),
                                  ['newlabel1', 'newlabel2'])

    def test_related_field_partial_update_no_permission(self):
        with self.object_permissions_mock_no_edit_labels:
            response = self.client.patch('/api/v2/objects/1/', {
                'name': 'newname',
                'labels': ['newlabel1', 'newlabel2'],
            }, format='json')

            self.assertEqual(response.status_code, 200)

            obj1 = InventoryObject.objects.get(id=1)

            self.assertEqual(obj1.name, 'newname')

            self.assertItemsEqual(obj1.label.values_list('name', flat=True),
                                  ['label1'])
