
"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""

from urllib import quote_plus

from django.contrib.auth.models import User
from django.test import TestCase, Client
from mock import patch

from inventory.models import (
    InventoryObject,
    IPv4Network,
    L2Network,
    Lab,
    Location,
    RacksUser,
    RacksGroup,
)
from security.permissions import Permissions


class WebUITestCase(TestCase):
    maxDiff = None

    def setUp(self):
        self.client = Client()

        InventoryObject.objects.create(
            id=1,
            name='object1',
        )

        L2Network.objects.create(id=1, vlanid=1)

        Lab.objects.create(id=1)

        Location.objects.create(id=1)

        IPv4Network.objects.create(id=1, address="1.2.3.4", netmask="25")

        self.urls = [
            '/',
            '/?label=mylabel',
            '/?responsible=blahuser',
            '/?label=mylabel&responsible',
            '/new',
            '/1',
            '/1/edit',

            '/vlans/',
            '/vlans/new',
            '/vlans/1/edit',

            '/labs/',
            '/labs/new',
            '/labs/1/edit',

            '/locations/',
            '/locations/new',
            '/locations/1/edit',

            '/ipv4networks/',
            '/ipv4networks/1',
            '/ipv4networks/1/edit',

            '/labels.html',
            '/search.html?query=blah',

            '/account/responsible',
            '/account/assigned',
        ]

    def test_authenticated_access(self):
        User.objects.create_superuser(username='testuser', password='testpassword',
                                      email='hello@world')

        self.client.login(username='testuser', password='testpassword')

        responses = {
            url: self.client.get(url)
            for url in self.urls
        }

        self.assertEqual({
            url: {
                'status_code': response.status_code,
                'content_type': response['Content-Type']
            }
            for url, response in responses.items()
        }, {
            url: {
                'status_code': 200,
                'content_type': 'text/html; charset=utf-8'
            }
            for url in self.urls
        })

    def test_unautenticated_access(self):
        responses = {
            url: self.client.get(url)
            for url in self.urls
        }

        self.assertEqual({
            url: {
                'status_code': response.status_code,
                'location': response.get('Location')
            }
            for url, response in responses.items()
        }, {
            url: {
                'status_code': 302,
                'location': (
                    'http://testserver/account/login.html?next=%s' %
                    quote_plus(url, safe='/')
                ),
            }
            for url in self.urls
        })


class NonPrivelegedUserTestCase(TestCase):
    maxDiff = None

    def setUp(self):
        User.objects.create_user(username='testuser', password='testpassword', email='hello@world')
        self.client = Client()
        self.client.login(username='testuser', password='testpassword')

        InventoryObject.objects.create(
            id=1,
            name='object1',
        )

        L2Network.objects.create(id=1, vlanid=1)

        Lab.objects.create(id=1)

        Location.objects.create(id=1)

        IPv4Network.objects.create(id=1, address="1.2.3.4", netmask="25")

        self.urls = [
            '/new',
            '/1/edit',
            '/1/delete',
            '/vlans/new',
            '/vlans/1/edit',
            '/vlans/1/delete',
            '/labs/new',
            '/labs/1/edit',
            '/labs/1/delete',
            '/locations/new',
            '/locations/1/edit',
            '/locations/1/delete',
        ]

    def test_get(self):
        # Non privileged user can open edit pages, to view all fields
        urls_with_status_codes = {
            '/new': 403,
            '/1/edit': 200,
            '/1/delete': 403,
            '/vlans/new': 403,
            '/vlans/1/edit': 200,
            '/vlans/1/delete': 403,
            '/labs/new': 403,
            '/labs/1/edit': 200,
            '/labs/1/delete': 403,
            '/locations/new': 403,
            '/locations/1/edit': 200,
            '/locations/1/delete': 403,
        }

        status_codes = {
            url: self.client.get(url).status_code for url in urls_with_status_codes
        }

        self.assertEqual(status_codes, urls_with_status_codes)

    def test_post(self):
        # All post requests are forbidden for non-privileged user
        urls = [
            '/new',
            '/1/edit',
            '/1/delete',
            '/vlans/new',
            '/vlans/1/edit',
            '/vlans/1/delete',
            '/labs/new',
            '/labs/1/edit',
            '/labs/1/delete',
            '/locations/new',
            '/locations/1/edit',
            '/locations/1/delete',
        ]
        status_codes = {
            url: self.client.post(url).status_code for url in urls
        }

        self.assertEqual(status_codes, {url: 403 for url in urls})


class HiddenValuesTestCase(TestCase):
    maxDiff = None

    url = '/1'

    def setUp(self):
        User.objects.create_user(username='testuser', password='testpassword', email='hello@world')
        User.objects.create_user(username='user2', password='testpassword', email='hello@world')
        User.objects.create_user(username='user3', password='testpassword', email='hello@world')
        User.objects.create_user(username='user4', password='testpassword', email='hello@world')
        User.objects.create_user(username='user5', password='testpassword', email='hello@world')
        User.objects.create_superuser(username='superuser', password='testpassword',
                                      email='hello@world')

        self.client = Client()

        user2 = RacksUser.objects.create(username='user2')
        group3 = RacksGroup.objects.create(name='group3')
        group3.users.create(username='user3')

        obj = InventoryObject.objects.create(
            id=1,
            name='name_value',
            ipmi_host='ipmi_host_value',
            ipmi_user='ipmi_user_value',
            ipmi_password='ipmi_password_value',
            assigned_user=user2,
            assigned_group=group3,
        )

        user4 = RacksUser.objects.create(username='user4')
        group5 = RacksGroup.objects.create(name='group5')
        group5.users.create(username='user5')

        obj.labs.create(
            assigned_user=user4,
            assigned_group=group5,
        )

    def test_hidden_fields_for_privileged_user(self):
        self.client.login(username='superuser', password='testpassword')

        self.assert_fields_visible('/1')
        self.assert_fields_visible('/1/edit')

    def test_hidden_fields_for_non_privileged_user(self):
        self.client.login(username='testuser', password='testpassword')

        self.assert_fields_hidden('/1')
        self.assert_fields_hidden('/1/edit')

    def test_hidden_fields_for_assigned_user(self):
        self.client.login(username='user2', password='testpassword')

        self.assert_fields_visible('/1')
        self.assert_fields_visible('/1/edit')

    def test_hidden_fields_for_assigned_group(self):
        self.client.login(username='user3', password='testpassword')

        self.assert_fields_visible('/1')
        self.assert_fields_visible('/1/edit')

    def test_hidden_fields_for_assigned_lab_user(self):
        self.client.login(username='user4', password='testpassword')

        self.assert_fields_visible('/1')
        self.assert_fields_visible('/1/edit')

    def test_hidden_fields_for_assigned_lab_group(self):
        self.client.login(username='user5', password='testpassword')

        self.assert_fields_visible('/1')
        self.assert_fields_visible('/1/edit')

    def assert_fields_visible(self, url):
        response = self.client.get(url)

        html = response.content

        # This should be visible for all users
        self.assertIn('name_value', html)

        self.assertIn('ipmi_host_value', html)
        self.assertIn('ipmi_user_value', html)
        self.assertIn('ipmi_password_value', html)

    def assert_fields_hidden(self, url):
        response = self.client.get(url)

        html = response.content

        # This should be visible for all users
        self.assertIn('name_value', html)

        self.assertNotIn('ipmi_host_value', html)
        self.assertNotIn('ipmi_user_value', html)
        self.assertNotIn('ipmi_password_value', html)


class PartialEditTestCase(TestCase):
    permissions_mock = patch('inventory.models.InventoryObject.get_permissions',
                             return_value=Permissions(edit=('dcim_id',)))

    def setUp(self):
        self.client = Client()

        User.objects.create_user(username='testuser', password='testpassword',
                                 email='hello@world')
        User.objects.create_superuser(username='testsuperuser', password='testpassword',
                                      email='hello@world')

        InventoryObject.objects.create(
            id=1,
            name='object1',
            dcim_id=1,
        )

    def test_unprivileged(self):
        with self.permissions_mock:
            self.client.login(username='testuser', password='testpassword')

            response = self.client.post('/1/edit', {'dcim_id': '2', 'name': 'newname'})

            self.assertEqual(response.status_code, 302)

            modified_object = InventoryObject.objects.get(id=1)

            self.assertEqual(modified_object.name, 'object1')
            self.assertEqual(modified_object.dcim_id, 2)

    def test_superuser(self):
        with self.permissions_mock:
            self.client.login(username='testsuperuser', password='testpassword')

            response = self.client.post('/1/edit', {'dcim_id': '2', 'name': 'newname'})

            self.assertEqual(response.status_code, 302)

            modified_object = InventoryObject.objects.get(id=1)

            self.assertEqual(modified_object.name, 'newname')
            self.assertEqual(modified_object.dcim_id, 2)
