from django import forms
from django.conf.urls import url
from django.contrib.auth.decorators import login_required
from django.core.paginator import Paginator
from django.core.urlresolvers import reverse
from django.forms.forms import BoundField
from django.forms.models import modelform_factory
from django.forms import widgets
from django.http import HttpResponseForbidden, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.views.generic import View
from django.utils.decorators import method_decorator, classonlymethod

from inventory.widgets import FilteredSelectMultiple
from security.permissions import ProtectedObject, get_permissions


class CRUDMixin(object):
    view_set = None
    model = None
    template_name = None

    @property
    def url_names(self):
        return self.view_set.url_names

    def get_model_verbose_name(self):
        return self.model._meta.verbose_name

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.permissions = get_permissions(request, self.model)

        if not self.view_allowed(self.permissions):
            return HttpResponseForbidden()

        return super(CRUDMixin, self).dispatch(request, *args, **kwargs)

    @classonlymethod
    def as_view(self, view_set):
        return super(CRUDMixin, self).as_view(view_set=view_set, model=view_set.model)

    def get_template_names(self):
        if self.template_name:
            return self.template_name

        path = self.view_set.prefix.replace('.', '/')
        template = self.template

        template = self.template

        return [
            '{}/{}.html'.format(path, template),
            'inventory/crud/{}.html'.format(template),
        ]

    def get_queryset(self):
        return self.model.objects.all()


class ListView(CRUDMixin, View):
    paginate_by = 20
    template = 'change_list'

    def view_allowed(self, permissions):
        return True

    def get(self, request):
        queryset = self.get_queryset()

        paginator = Paginator(queryset, self.paginate_by)
        page = paginator.page(request.GET.get('page', 1))

        context = {
            'view': self,
            'is_paginated': True,
            'page_obj': page,
            'paginator': paginator,
            'object_list': page.object_list,
            'permissions': self.permissions,
        }

        return render(request, self.get_template_names(), context)


class DetailView(CRUDMixin, View):
    template = 'detail'

    def view_allowed(self, permissions):
        return True

    def get(self, request, pk):
        self.object = get_object_or_404(self.get_queryset(), pk=pk)

        self.permissions = get_permissions(request, self.model, self.object)
        self.protected_object = ProtectedObject(self.object, self.permissions)

        context = {
            'view': self,
            'object': self.protected_object,
        }

        if hasattr(self, 'get_context'):
            context.update(self.get_context())

        return render(request, self.get_template_names(), context)


class FieldForDisplay(object):
    def __init__(self, label):
        self.label = label


class ReadOnlyBoundField(BoundField):
    def as_widget(self, widget=None, attrs=None, only_initial=False):
        attrs = attrs or {}
        attrs['readonly'] = 'readonly'

        widget = widget or self.field.widget

        # Disable opening of Select fields. We're checking for type, because our
        # FilteredSelectMultiple widget is subclass of Select, but correctly handles readonly
        # attribute.
        if type(widget) is widgets.Select:
            attrs['disabled'] = True

        return super(ReadOnlyBoundField, self).as_widget(widget, attrs, only_initial)


class CustomizedFormMixin(object):
    def get_initial(self, obj):
        initial = {}

        for field_name in self.view_set.related_foreign_key_fields:
            if obj:
                initial[field_name] = getattr(obj, field_name).all()

        return initial

    def get_form_class(self):
        if self.view_set.change_form_class:
            return self.view_set.change_form_class

        widgets = {
            field: FilteredSelectMultiple
            for field in self.view_set.filtered_select_fields
        }

        if self.view_set.widgets:
            widgets.update(self.view_set.widgets)

        return modelform_factory(self.model, fields=self.view_set.fields, widgets=widgets,
                                 formfield_callback=self.formfield_callback)

    def get_fields_for_form_display(self, form):
        editable_fields = {}
        fields_for_display = []

        for field_name, field in form.fields.items():
            bound_field = form[field_name]
            field_for_display = FieldForDisplay(bound_field.label)
            field_permissions = self.permissions.get_field_permissions(field_name)

            editable = (field_permissions.edit and field_name not in self.view_set.readonly_fields)
            readable = field_permissions.view

            field_for_display.form_field = bound_field

            if editable:
                editable_fields[field_name] = field
            elif readable:
                field_for_display.form_field = ReadOnlyBoundField(form, field, field_name)
            else:
                field_for_display.no_permission = True

            fields_for_display.append(field_for_display)

        return fields_for_display, editable_fields

    def get_form(self, obj=None, data=None):
        form_class = self.get_form_class()
        form = form_class(data=data, instance=obj, initial=self.get_initial(obj))

        for field_name in self.view_set.related_foreign_key_fields:
            field = getattr(self.model, field_name)

            qs = field.related.model.objects.all()

            form.fields[field_name] = forms.ModelMultipleChoiceField(qs,
                                                                     required=False,
                                                                     widget=FilteredSelectMultiple)

        form.fields_for_display, form.fields = self.get_fields_for_form_display(form)

        # If no fields are editable, disable save
        form.allow_save = bool(form.fields)

        return form

    def form_valid(self, form):
        response = super(CustomizedFormMixin, self).form_valid(form)

        for field_name in self.view_set.related_foreign_key_fields:
            value = form.cleaned_data[field_name]

            setattr(self.object, field_name, value)

        return response

    def formfield_callback(self, f, **kwargs):
        field_class = self.view_set.field_classes.get(f.name)

        if field_class:
            return field_class(**kwargs)

        return f.formfield(**kwargs)


class UpdateView(CRUDMixin, CustomizedFormMixin, View):
    template_name = 'inventory/crud/update_form.html'

    def view_allowed(self, permissions):
        return True

    def get_success_url(self, obj):
        try:
            return obj.get_absolute_url()
        except AttributeError:
            return reverse(self.url_names['list'])

    def get(self, request, pk):
        obj = get_object_or_404(self.get_queryset(), pk=pk)

        self.permissions = get_permissions(request, self.model, obj)

        form = self.get_form(obj=obj)

        context = {
            'object': ProtectedObject(obj, self.permissions),
            'form': form,
        }

        return render(request, self.get_template_names(), context)

    def post(self, request, pk):
        obj = get_object_or_404(self.get_queryset(), pk=pk)

        form = self.get_form(obj=obj, data=request.POST)

        if not form.allow_save:
            return HttpResponseForbidden()

        if form.is_valid():
            obj = form.save()
            return HttpResponseRedirect(self.get_success_url(obj))

        context = {
            'form': form,
        }

        return render(request, self.get_template_names(), context)


class CreateView(CRUDMixin, CustomizedFormMixin, View):
    template_name = 'inventory/crud/create_form.html'

    def view_allowed(self, permissions):
        return permissions.create

    def get_success_url(self, new_object):
        try:
            new_object.get_absolute_url()
        except AttributeError:
            return reverse(self.url_names['list'])

    def get(self, request):
        form = self.get_form()

        context = {
            'form': form,
        }

        return render(request, self.get_template_names(), context)

    def post(self, request):
        form = self.get_form(data=request.POST)

        if not form.allow_save:
            return HttpResponseForbidden()

        if form.is_valid():
            new_object = form.save()
            return HttpResponseRedirect(self.get_success_url(new_object))

        context = {
            'form': form,
        }

        return render(request, self.get_template_names(), context)


class DeleteView(CRUDMixin, View):
    template_name = 'inventory/crud/confirm_delete.html'

    def view_allowed(self, permissions):
        return permissions.delete

    def get_success_url(self):
        return self.request.GET.get('next') or reverse(self.router.list_view.url_name)

    def post(self, request, pk):
        obj = get_object_or_404(self.get_queryset(), pk=pk)

        obj.delete()

        return HttpResponseRedirect(self.get_success_url())


class ViewSet(object):
    PATTERNS = (
        (r'^$', 'list'),
        (r'^(?P<pk>[0-9]+)$', 'detail'),
        (r'^new$', 'create'),
        (r'^(?P<pk>[0-9]+)/edit$', 'update'),
        (r'^(?P<pk>[0-9]+)/delete$', 'delete'),
    )

    list_view = ListView
    detail_view = None
    create_view = CreateView
    update_view = UpdateView
    delete_view = DeleteView

    change_form_class = None
    fields = None
    filtered_select_fields = ()
    readonly_fields = ()
    related_foreign_key_fields = ()

    fields = None
    field_classes = {}
    widgets = None

    def __init__(self, prefix):
        self.prefix = prefix
        self.url_names = {}
        self.urlpatterns = []

        for pattern, name in self.PATTERNS:
            view_attr = name + '_view'
            view_class = getattr(self, view_attr)

            if view_class:
                url_name = '%s.%s' % (prefix, name)
                self.url_names[name] = url_name
                self.urlpatterns.append(url(pattern, view_class.as_view(self), name=url_name))

            self.url_names['object'] = self.url_names.get('detail') or self.url_names.get('update')
