from django.utils.html import mark_safe

from inventory.forms import InventoryObjectForm
from inventory.models import InventoryObject
from inventory.views import crud
from security import permissions


class InventoryObjectListView(crud.ListView):
    def get_queryset(self):
        filters = self.get_filters()

        return InventoryObject.objects.select_related(
            'facter_location', 'facter_location__location', 'type').filter(**filters)

    def get_filters(self):
        request = self.request

        label = request.GET.get('label')
        responsible = request.GET.get('responsible')

        filters = {}

        if label is not None:
            if label == 'None':
                filters['label'] = None
            else:
                filters['label__name'] = label

        if responsible is not None:
            if responsible == 'None':
                filters['responsible'] = None
            else:
                filters['responsible__username'] = responsible

        return filters


class InventoryObjectDetailView(crud.DetailView):
    template_name = 'inventory/objects/detail.html'

    MODEL_FIELD_ATTRIBUTES = ['ipmi_host', 'ipmi_user', 'ipmi_password']

    def get_context(self, **kwargs):
        context = {
            'attributes': self.get_attributes()
        }

        protected = self.protected_object

        if protected.ssh_link:
            if permissions.can_view(protected.ssh_link):
                context['ssh_link'] = protected.ssh_link

        if protected.ipmi_link and permissions.can_view(protected.ipmi_link):
            if protected.ipmi_user and protected.ipmi_password:
                if permissions.can_view(protected.ipmi_user, protected.ipmi_password):
                    context['ipmi_form'] = {
                        'action': protected.ipmi_link + '/cgi/login.cgi',
                        'name': protected.ipmi_user,
                        'pwd': protected.ipmi_password,
                    }
            else:
                context['ipmi_link'] = protected.ipmi_link

        return context

    def get_attributes(self):
        obj = self.protected_object

        attributes = [(attr.name, mark_safe(attr.awesome_value())) for attr in obj.attributes.all()]

        for attr_name in self.MODEL_FIELD_ATTRIBUTES:
            protected_value = getattr(obj, attr_name)

            if protected_value is not None:
                attributes.append((attr_name, protected_value))

        return attributes


class InventoryObjectViewSet(crud.ViewSet):
    model = InventoryObject
    change_form_class = InventoryObjectForm
    list_view = InventoryObjectListView
    detail_view = InventoryObjectDetailView

    readonly_fields = ('orphan_dcim_id',)


urlpatterns = InventoryObjectViewSet('inventory.objects').urlpatterns
