from launchpadlib.launchpad import Launchpad, AnonymousAccessToken
from launchpadlib.credentials import Credentials, AccessToken
from lazr.restfulclient.errors import HTTPError


class AuthorizationEngine(object):
    # Custom authorization engine, reverts to anonymous access when we get Unauthorized error
    # (due to missing or revoked access token)
    def __init__(self, request):
        self.request = request

    def __call__(self, credentials, credential_store):
        credentials.access_token = AnonymousAccessToken()
        return credentials


class LaunchpadAuthorizationMiddleware:
    def process_request(self, request):
        credentials = Credentials("Racks App")
        request.launchpad = Launchpad(credentials, AuthorizationEngine(request), None,
                                      service_root='production')

        request_token_string = request.COOKIES.get('launchpadrequesttoken')
        access_token_string = request.COOKIES.get('launchpadaccesstoken')

        if access_token_string:
            # Use access token from cookie
            credentials.access_token = AccessToken.from_string(access_token_string)
        elif request_token_string:
            # Use request token from cookie, so we could finish authorization process and get
            # access token
            credentials._request_token = AccessToken.from_string(request_token_string)
            try:
                credentials.exchange_request_token_for_access_token(web_root='production')
            except HTTPError:
                pass

    def process_response(self, request, response):
        access_token = request.launchpad.credentials.access_token
        request_token = request.launchpad.credentials._request_token

        # Access token is not valid anymore due to some reason, delete it from cookies
        if access_token is None or isinstance(access_token, AnonymousAccessToken):
            response.delete_cookie('launchpadaccesstoken')
            access_token = None

        if access_token is not None:
            # Got valid access token, store it in cookie
            response.set_cookie('launchpadaccesstoken', access_token.to_string())
            # We don't need request token anymore
            response.delete_cookie('launchpadrequesttoken')
        elif request_token:
            # Save request token to cookie, so we could use it for authorization on next request to
            # Racks
            response.set_cookie('launchpadrequesttoken', request_token.to_string())

        return response
