import json

from django.conf import settings
from django.http import HttpResponse, HttpResponseRedirect

from launchpadlib.launchpad import Launchpad
from launchpadlib.credentials import Credentials, AccessToken, AnonymousAccessToken
from lazr.restfulclient.errors import HTTPError


CONSUMER_NAME = "Racks App"


def authorize(request):
    credentials = Credentials(CONSUMER_NAME)
    authorization_url = credentials.get_request_token(web_root=settings.LAUNCHPAD_SERVICE_ROOT)
    request_token = credentials._request_token

    redirect_url = request.GET.get('next', '/')
    authorization_url += '&oauth_callback=' + redirect_url

    response = HttpResponseRedirect(authorization_url)
    response.set_cookie(COOKIE_NAMES.request_token, request_token.to_string())
    return response


def authorization_engine(credentials, credential_store):
    credentials.access_token = AnonymousAccessToken()
    # Must return credentials passed
    return credentials


def get_credentials(request):
    request_token_string = request.COOKIES.get('launchpadrequesttoken')
    access_token_string = request.COOKIES.get('launchpadaccesstoken')

    request_token = (AccessToken.from_string(request_token_string)
                     if request_token_string else None)
    access_token = (AccessToken.from_string(access_token_string)
                    if access_token_string else None)

    token = AnonymousAccessToken()

    if access_token:
        # Use access token from cookie
        # credentials = Credentials(CONSUMER_NAME, access_token=token)
        token = access_token
    elif request_token:
        # Use request token from cookie, so we could finish authorization process and get
        # access token
        credentials = Credentials(CONSUMER_NAME)
        credentials._request_token = request_token

        try:
            credentials.exchange_request_token_for_access_token(
                web_root=settings.LAUNCHPAD_SERVICE_ROOT)
        except HTTPError, e:
            if e.response.status == 403:
                # Request token is no longer valid
                pass
            elif e.response.status == 401:
                # Request is not authorized yet
                pass
            else:
                raise
        else:
            # Success, use the credentials
            return credentials

    return Credentials(CONSUMER_NAME, access_token=token)


def fetch_bugs(launchpad, bugs_ids):
    res = {}

    for bug_id in bugs_ids:
        try:
            bug = launchpad.bugs[bug_id]
        except KeyError:
            data = {
                'error': 'Bug not found'
            }
        else:
            data = {
                'id': bug.id,
                'title': bug.title,
                'webLink': bug.web_link,
            }

        res[bug_id] = data

    return res


class COOKIE_NAMES:
    authorization_url = 'launchpadauthorizationurl'
    access_token = 'launchpadaccesstoken'
    request_token = 'launchpadrequesttoken'


def fetch_bugs_json(request):
    credentials = get_credentials(request)

    launchpad = Launchpad(credentials, authorization_engine, None,
                          service_root=settings.LAUNCHPAD_SERVICE_ROOT,
                          cache=settings.LAUNCHPAD_CACHEDIR)

    bugs_ids = request.GET.getlist('id')
    res = {
        'bugs': fetch_bugs(launchpad, bugs_ids)
    }

    # This check should be performed after actual bug fetch, because revoked token will reset
    # only after unsuccessful attempt to use it
    access_token = launchpad.credentials.access_token
    if isinstance(access_token, AnonymousAccessToken):
        # Anonymous access token is not real
        access_token = None
    if access_token:
        res['authorized'] = True

    content = json.dumps(res)
    response = HttpResponse(content, content_type='application/json')

    if access_token:
        # Save valid access token
        response.set_cookie(COOKIE_NAMES.access_token, access_token.to_string())
        # Request token is not needed anymore
        response.delete_cookie(COOKIE_NAMES.request_token)
    else:
        response.delete_cookie(COOKIE_NAMES.access_token)

    return response
