"""
Django settings for racks project.

For more information on this file, see
https://docs.djangoproject.com/en/1.6/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.6/ref/settings/
"""

# Build paths inside the project like this: os.path.join(PROJECT_ROOT, ...)
import os

from django.contrib.messages import constants as messages

import site_settings


PROJECT_ROOT = os.path.dirname(os.path.dirname(__file__))

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/var/www/example.com/static/"
STATIC_ROOT = ''

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = []

# Twitter Bootstrap compability
MESSAGE_TAGS = {
    messages.ERROR: 'danger',
}

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.6/howto/deployment/checklist/

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'racks',

    'south',
    'rest_framework',
    'bootstrap3',

    'inventory',
    'dcim',
    'importers',  # Celery tasks

    'reports',
    'security',
    'account',
]

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
)

NETWORK_SWITCHES_CREDENTIALS_PATH = 'network-switches-credentials.yaml'

# Celery tasks hard time limits, in seconds
TASK_TIME_LIMITS = {
    # Zabbix can be slow
    'importers.tasks.import_zabbix_hosts': 5 * 60,
    'importers.tasks.import_puppet_attributes': 5 * 60,
    # DCIM can be slow
    'importers.tasks.import_dcim_servers': 5 * 60,
    'importers.tasks.import_dcim_switches': 5 * 60,
    # SSH to each switch, could take longer than other
    'importers.tasks.import_switches_address_tables': 10 * 60,
    # Internal work with database, should be fast
    'importers.tasks.bind_dcim_to_switch_ports': 1 * 60,
    'importers.tasks.bind_dcim_servers_to_inventory': 1 * 60,
    'importers.tasks.import_users_from_ldap': 5 * 60,
}

LAUNCHPAD_CACHEDIR = '/tmp/racks/launchpadlib/cache/'
LAUNCHPAD_SERVICE_ROOT = 'production'

DCIM_JIRA_ISSUE_CUSTOM_VALUE_ID = None

PUPPET = None

LOGIN_URL = '/account/login.html'
LOGOUT_URL = '/account/logout.html'
LOGIN_REDIRECT_URL = '/'

REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    # or allow read-only access for unauthenticated users.
    'DEFAULT_PERMISSION_CLASSES': (
        'security.rest_api.Permissions',
    ),
    'DEFAULT_FILTER_BACKENDS': ('rest_framework.filters.DjangoFilterBackend',),
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
        'security.authentication.ApplicationAuthentication',
    ),
}

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'racks.navigation.NavigationMiddleware',
    'security.middleware.RacksUserMiddleware',
)

TEMPLATE_LOADERS = (
    'django.template.loaders.app_directories.Loader',
)

TEMPLATE_CONTEXT_PROCESSORS = (
    "django.contrib.auth.context_processors.auth",
    "django.core.context_processors.request",
    "django.core.context_processors.debug",
    "django.core.context_processors.i18n",
    "django.core.context_processors.media",
    "django.core.context_processors.static",
    "django.core.context_processors.tz",
    "django.contrib.messages.context_processors.messages",
)

ROOT_URLCONF = 'racks.urls'

WSGI_APPLICATION = 'racks.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.6/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_ROOT, 'db.sqlite3'),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.6/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s %(levelname)-8s %(name)-15s %(message)s',
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        },
        'verbose console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        },
    },
    'loggers': {
        'django': {
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'INFO'),
        },
        'importers': {
            'handlers': ['verbose console'],
            'level': 'INFO',
        },
        'django_auth_ldap': {
            'handlers': ['console'],
            'level': 'DEBUG',
        }
    },
}

site_settings.update_settings(globals(), 'RACKSCONFPATH', '.:/etc/racks')
