(function($) {
  function Fetcher() {
    this.cache = {};
  }

  Fetcher.prototype = {
    cache: {},

    fetch: function(bugs) {
      var _this = this;
      var toFetch = [];

      $.each(bugs, function() {
        var bug = this;

        if (!(bug.id in _this.cache)) {
          toFetch.push(bug.id);
          _this.cache[bug.id] = {loading: true};
          if (bug.fetchStart) {
            bug.fetchStart();
          }
        }
      });

      if (toFetch.length) {
        $.ajax({
          url: '/launchpad/fetchbugs',
          data: {id: toFetch},
          traditional: true,
          dataType: 'json',
          success: function(res) {
            if (!res.authorized) {
              $('#launchpad-authorization-link').removeClass('hidden');
            }
            $.extend(_this.cache, res.bugs);
            _this.onFetchSuccess(bugs);
          }
        });
      } else {
        this.onFetchSuccess(bugs);
      }
    },

    onFetchSuccess: function(bugs) {
      var _this = this;

      $.each(bugs, function() {
        var bug = this;

        if (bug.id in _this.cache) {
          bug.fetchSuccess(_this.cache[bug.id]);
        }
      });
    }
  };

  function fillBugTitle(titleContainer, data) {
    titleContainer.empty();

    if (data.error) {
      $('<span class="text-danger"></span>').text(data.error).appendTo(titleContainer);
    } else {
      $('<a target="_blank"></a>')
        .text(data.title)
        .attr('href', data.webLink)
        .appendTo(titleContainer);
    }
  }

  function BugInfo(elem) {
    this.id = elem.attr('data-fetch-bug');
    this.elem = elem;
  }

  BugInfo.prototype = {
    fetchStart: function() {
      this.titleContainer = $('<span>Loading...</span>').appendTo(this.elem);
    },
    fetchSuccess: function(data) {
      fillBugTitle(this.titleContainer, data);
    }
  };

  var fetcher = new Fetcher();

  fetcher.fetch($('[data-fetch-bug]').map(function() {
    return new BugInfo($(this));
  }));

  function BugsInputEntry(id, control) {
    this.id = id;
    this.control = control;
    this.construct();
  }

  BugsInputEntry.prototype = {
    construct: function() {
      var _this = this;
      var entry = $('<div>');

      entry.attr('data-bug-id', this.id);

      $('<span>').text('LP#' + this.id).appendTo(entry);
      entry.append(' ');
      this.titleContainer = $('<span>').appendTo(entry);

      if (!this.control.readonly) {
        entry.append(' ');
        this.removeLink = $('<span class="btn btn-xs btn-link"></span>')
          .appendTo(entry)
          .html('<span class="text-danger glyphicon glyphicon-remove"></span>')
          .click(entry, function(e) {
            _this.control.onListEntryRemoveClick(e);
          });
      }

      this.elem = entry;
    },
    fetchStart: function() {
      this.titleContainer.html('Loading...</span>');
    },
    fetchSuccess: function(data) {
      fillBugTitle(this.titleContainer, data);
    }
  };

  function BugsInput(input) {
    this.hiddenInput = input;

    this.readonly = input.attr('readonly');

    var val = input.val();
    var ids = val.length ?  input.val().split(',') : [];

    this.control = this.constructControl(ids);

    input.replaceWith(this.control);
    this.control.append(input.attr('type', 'hidden'));
  }

  BugsInput.prototype = {
    val: function(val) {
      return this.list.children('div').map(function() {
        return $(this).attr('data-bug-id');
      }).toArray();
    },

    updateHiddenInput: function() {
      var ids = this.val();
      this.hiddenInput.val(ids.join(','));
    },

    constructControl: function(ids) {
      var _this = this;

      var elem = $('<div>');
      this.list = $('<div>').appendTo(elem);

      if (!this.readonly) {
        this.openFormButton =
        $('<span class="form-control btn btn-default">Add bug</span>')
          .appendTo(elem).click(_this.onOpenFormButtonClick.bind(_this));
      }

      var entries = ids.map(function(id) {
        var entry = new BugsInputEntry(id, _this);
        _this.list.append(entry.elem);
        return entry;
      });

      fetcher.fetch(entries);

      return elem;
    },

    onOpenFormButtonClick: function() {
      this.openFormButton.addClass('hidden');
      this.form = this.constructForm();
      this.control.append(this.form);
    },

    removeForm: function() {
      this.form.remove();
      this.openFormButton.removeClass('hidden');
    },

    onListEntryRemoveClick: function(e) {
      e.data.remove();
      this.updateHiddenInput();
    },

    constructForm: function() {
      var elem = $('<div class="form">');

      var group1 = $('<div class="form-group"></div>').appendTo(elem);
      this.idInput = $('<input class="form-control" type="number">')
        .appendTo(group1)
        .on('change keyup', this.onIdInputChange.bind(this));
      this.helpBlock = $('<span class="help-block"></span>').appendTo(group1);

      if (!this.readonly) {
        var group2 = $('<div class="form-group"></div>').appendTo(elem);
        this.addButton = $('<span class="btn btn-default" disabled>Add</span>')
          .appendTo(group2)
          .click(this.onAddButtonClick.bind(this));

        group2.append(' ');
        $('<span class="btn btn-default">Cancel</span>')
          .appendTo(group2)
          .click(this.removeForm.bind(this));
      }

      return elem;
    },

    onAddButtonClick: function() {
      var id = this.idInput.val();
      var entry = new BugsInputEntry(id, this);
      this.list.append(entry.elem);
      fetcher.fetch([entry]);
      this.removeForm();
      this.updateHiddenInput();
    },

    onIdInputChange: function() {
      var id = this.idInput.val();

      if (!id.length) {
        this.onIdInputError('Please enter bug id');
        return;
      }

      if (this.val().indexOf(id) !== -1) {
        this.onIdInputError('This bug already associated');
        return;
      }

      this.form.removeClass('has-success has-error');
      this.addButton.attr('disabled', true);

      var _this = this;

      var bug = {
        id: id,
        fetchSuccess: function(data) {
          _this.onIdInputBugFetchSuccess(this.id, data);
        }
      };

      this.helpBlock.text('Loading...');

      fetcher.fetch([bug]);
    },

    onIdInputBugFetchSuccess: function(id, data) {
      var val = this.form ? this.idInput.val() : null;

      if (val !== null && val == id) {
        if (data.error) {
          this.onIdInputError(data.error);
        } else {
          this.onIdInputSuccess(data);
        }
      }
    },

    onIdInputError: function(text) {
      this.form.removeClass('has-success').addClass('has-error');
      this.helpBlock.text(text);
      this.addButton.attr('disabled', true);
    },

    onIdInputSuccess: function(data) {
      fillBugTitle(this.helpBlock, data);
      this.form.removeClass('has-error').addClass('has-success');
      this.addButton.attr('disabled', false);
    }
  };

  $.fn.bugsinput = function() {
    this.each(function() {
      new BugsInput($(this));
    });
  };

  $('[data-role=bugsinput]').bugsinput();
})(jQuery);
