class FieldPermissions(object):
    def __init__(self, edit=False, view=False):
        self.edit = edit
        self.view = view


FIELD_EDIT = FieldPermissions(edit=True, view=True)
FIELD_VIEW = FieldPermissions(view=True)
FIELD_NONE = FieldPermissions()


class EditAll(object):
    def __init__(self, create=False, delete=False):
        self.create = create
        self.delete = delete

    def get_field_permissions(self, field_name):
        return FIELD_EDIT


ABSOLUTE = EditAll(create=True, delete=True)


class ViewAll(object):
    create = delete = False

    def get_field_permissions(self, field_name):
        return FIELD_VIEW


VIEW_ALL = ViewAll()


class Permissions(object):
    def __init__(self, create=False, delete=False, edit=(), view=()):
        self.create = create
        self.delete = delete
        self.edit = edit
        self.view = view

    def get_field_permissions(self, field_name):
        if field_name in self.edit:
            return FIELD_EDIT
        if field_name in self.view:
            return FIELD_VIEW

        return FIELD_NONE


class ValueWithoutAccess(object):
    # Attributes have no access too
    def __getattr__(self, name):
        return self

    # Result of calls have no access either
    def __call__(self, *args, **kwargs):
        return self

    def __iter__(self):
        return iter([self])

    def __len__(self):
        return 1


VALUE_WITHOUT_ACCESS = ValueWithoutAccess()


class PermissionsAccessor(object):
    """Class for testing field permissions in templates"""

    def __init__(self, permissions):
        self.__permissions = permissions

    def __getattr__(self, name):
        return self.__permissions.get_field_permissions(name)


class ProtectedObject(object):
    def __init__(self, obj, permissions):
        self.__object = obj
        self.__permissions = permissions

        self.permissions = PermissionsAccessor(permissions)

    def __getattr__(self, name):
        value = getattr(self.__object, name)

        # Special case for links, should be always accessible
        if name == 'pk':
            return value

        # Special case for None, considered safe value
        if value is None:
            return None

        # Special case for blank, considered safe value
        if value == '':
            return ''

        field_permissions = self.__permissions.get_field_permissions(name)

        if field_permissions.view:
            return value

        return VALUE_WITHOUT_ACCESS


def can_view(*values):
    for value in values:
        if value is VALUE_WITHOUT_ACCESS:
            return False

    return True


def get_permissions(request, model, obj=None):
    user = request.user

    if user and user.is_superuser:
        return ABSOLUTE

    if hasattr(model, 'get_permissions'):
        # Do it here to avoid circular imports
        from inventory.models import RacksUser
        racks_user = None
        if user:
            try:
                racks_user = RacksUser.objects.get(username=user.username)
            except RacksUser.DoesNotExist:
                pass
        return model.get_permissions(racks_user, obj)

    # Default permissions
    return VIEW_ALL
