import hashlib
import random

from django.contrib import messages
from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.decorators import login_required
from django.db import IntegrityError
from django.shortcuts import redirect
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _

from security.models import Application


@login_required
def applications(request):
    applications = Application.objects.filter(
        user=request.user).order_by('-active')
    if request.method == 'POST':
        action = request.POST.get('action')
        if action == 'create':
            name = request.POST.get('name')
            if name == '':
                messages.add_message(
                    request,
                    messages.ERROR,
                    _('Wrong application name: "%s"' % (name)))
                return redirect('/account/applications.html')

            t = hashlib.sha512()
            t.update(request.user.username)
            t.update(name)
            t.update(str(random.random()))

            try:
                application, application_created = \
                    Application.objects.get_or_create(
                        name=name,
                        auth_token=t.hexdigest(),
                        user=request.user)
            except IntegrityError:
                messages.add_message(
                    request,
                    messages.ERROR,
                    _('You already have application named "%s"' % (name)))
                return redirect('/account/applications.html')

            if application_created:
                application.save()

            messages.add_message(
                request,
                messages.SUCCESS,
                _('Application "%s" created successfully' % (name)))
            return redirect('/account/applications.html')
        elif action == 'revoke':
            app = Application.objects.get(pk=request.POST.get('pk'))
            app.active = False
            app.save()

            messages.add_message(
                request,
                messages.SUCCESS,
                _('Application "%s" successfully revoked' % (app.name)))
            return redirect(
                request.POST.get('redirect', '/account/applications.html'))
        elif action == 'restore':
            app = Application.objects.get(pk=request.POST.get('pk'))
            app.active = True
            app.save()

            messages.add_message(
                request,
                messages.SUCCESS,
                _('Application "%s" successfully restored' % (app.name)))
            return redirect(
                request.POST.get('redirect', '/account/applications.html'))
    else:
        return render_to_response('security/applications.html', {
            'applications': applications,
            'subsection_applications': True,
        }, context_instance=RequestContext(request))


@login_required
def application(request, applicationid):
    application = Application.objects.get(pk=applicationid)
    return render_to_response('security/application.html', {
        'application': application,
        'subsection_applications': True,
    }, context_instance=RequestContext(request))


def login_view(request):
    redirect_url = request.GET.get('next', '/')

    if request.user.is_authenticated():
        return redirect(redirect_url)

    if request.method == 'POST':
        username = request.POST['username']
        password = request.POST['password']
        user = authenticate(username=username, password=password)
        if user is not None:
            if user.is_active:
                login(request, user)
                return redirect(redirect_url)
            else:
                return render_to_response('security/login.html', {
                    'message': _(
                        'Invalid credentials: account is deactivated'),
                    'redirect_url': redirect_url,
                }, context_instance=RequestContext(request))

        return render_to_response('security/login.html', {
            'message': _('Invalid credentials'),
            'redirect_url': redirect_url,
        }, context_instance=RequestContext(request))
    else:
        return render_to_response('security/login.html', {
            'redirect_url': redirect_url,
        }, context_instance=RequestContext(request))


def logout_view(request):
    redirect_url = request.GET.get('next', '/')

    logout(request)

    return redirect(redirect_url)


@login_required
def profile(request):
    return render_to_response('security/profile.html', {
        'subsection_profile': True,
    }, context_instance=RequestContext(request))
