from __future__ import absolute_import

import os

import yaml
from celery.schedules import crontab
from django.core.exceptions import ImproperlyConfigured


def load_yaml(filename, conf_dirs):
    for conf_dir in conf_dirs:
        path = os.path.join(conf_dir, filename)

        if os.path.exists(path):
            with open(path) as f:
                return yaml.safe_load(f)

    raise ImproperlyConfigured('%s not found in %r' % (filename, conf_dirs))


def process_ldap_search_dict(settings, name):
    try:
        search = settings[name].copy()
    except KeyError:
        pass
    else:
        import ldap
        from django_auth_ldap.config import LDAPSearch

        if 'scope' in search:
            search['scope'] = getattr(ldap, search['scope'])

        settings[name] = LDAPSearch(**search)


def process_ldap_settings_dict(settings):
    if 'AUTH_LDAP_SERVER_URI' not in settings:
        return

    from django_auth_ldap.config import GroupOfNamesType

    settings['AUTH_LDAP_GROUP_TYPE'] = GroupOfNamesType()
    if not settings.get('AUTH_LDAP_DISABLE_BACKEND'):
        settings['AUTHENTICATION_BACKENDS'] = ('django_auth_ldap.backend.LDAPBackend',)
    settings['AUTH_LDAP_MIRROR_GROUPS'] = True

    process_ldap_search_dict(settings, 'AUTH_LDAP_USER_SEARCH')
    process_ldap_search_dict(settings, 'AUTH_LDAP_GROUP_SEARCH')

    process_ldap_search_dict(settings, 'LDAP_USER_SEARCH')
    process_ldap_search_dict(settings, 'LDAP_GROUP_SEARCH')


def process_celery_settings(settings):
    if 'CELERYBEAT_SCHEDULE' in settings:
        for params in settings['CELERYBEAT_SCHEDULE'].values():
            if 'crontab' in params['schedule']:
                params['schedule'] = crontab(**params['schedule']['crontab'])


def update_settings(settings, conf_path_envvar, conf_path_default, filename="settings.yaml"):
    conf_path = os.getenv(conf_path_envvar, conf_path_default)
    conf_dirs = settings['SITE_CONF_DIRS'] = conf_path.split(':')

    _site_settings_dict = load_yaml(filename, conf_dirs)
    process_ldap_settings_dict(_site_settings_dict)
    process_celery_settings(_site_settings_dict)

    settings.update(_site_settings_dict)
